"""
Basic GridCal (4.6.1) interface, based on the pandapower interface written by Jinning Wang

Author: Josep Fanals (@JosepFanals)
"""

import logging
from functools import wraps

import numpy as np
import matplotlib

from andes.shared import GridCal_Engine as gc

matplotlib.use('agg')
logger = logging.getLogger(__name__)


def require_gridcal(f):
    """
    Decorator for functions that require GridCal.
    """

    @wraps(f)
    def wrapper(*args, **kwds):
        try:
            getattr(gc, '__name__')
        except AttributeError as exc:
            raise ModuleNotFoundError("GridCal needs to be installed manually.") from exc

        return f(*args, **kwds)

    return wrapper


def _to_gc_bus(ssp, ssa_bus):
    """
    Define buses in GridCal's grid.
    """

    dic_bus = {}

    for i in range(ssa_bus.n):
        bus = gc.Bus(name=ssa_bus.name.v[i],
                     active=ssa_bus.u.v[i],
                     vnom=ssa_bus.Vn.v[i],
                     vmin=ssa_bus.vmin.v[i],
                     vmax=ssa_bus.vmax.v[i],
                     xpos=ssa_bus.xcoord.v[i],
                     ypos=ssa_bus.ycoord.v[i],
                     Vm0=ssa_bus.v0.v[i],
                     Va0=ssa_bus.a0.v[i])

        dic_bus[ssa_bus.idx.v[i]] = bus

        ssp.add_bus(bus)

    return ssp, dic_bus


def _to_gc_branch(ssp, ssa_line, dic_bus):
    """
    Define branches (lines and trafos) in GridCal's grid.
    """

    rate = [xx if xx != 0.0 else 1.0 for xx in ssa_line.rate_a.v]  # 1.0 is the default in GridCal

    for i in range(len(ssa_line)):

        if ssa_line.tap.v[i] != 1.0 or ssa_line.phi.v[i] != 0.0:
            # trafo
            trafo = gc.Transformer2W(bus_from=dic_bus[ssa_line.bus1.v[i]],
                                     bus_to=dic_bus[ssa_line.bus2.v[i]],
                                     name=ssa_line.name.v[i],
                                     active=ssa_line.u.v[i],
                                     r=ssa_line.r.v[i],
                                     x=ssa_line.x.v[i],
                                     b=ssa_line.b.v[i],
                                     g=ssa_line.g.v[i],
                                     rate=rate[i],
                                     tap=ssa_line.tap.v[i],
                                     shift_angle=ssa_line.phi.v[i])

            ssp.add_transformer2w(trafo)

        else:
            # line
            line = gc.Line(bus_from=dic_bus[ssa_line.bus1.v[i]],
                           bus_to=dic_bus[ssa_line.bus2.v[i]],
                           name=ssa_line.name.v[i],
                           active=ssa_line.u.v[i],
                           r=ssa_line.r.v[i],
                           x=ssa_line.x.v[i],
                           b=ssa_line.b.v[i],
                           rate=rate[i])

            ssp.add_line(line)

    return ssp


def _to_gc_load(ssp, ssa_load, dic_bus, sbase=1.0):
    """
    Define loads in GridCal's grid.
    """

    for i in range(len(ssa_load)):
        load = gc.Load(name=ssa_load.name.v[i],
                       active=ssa_load.u.v[i],
                       P=ssa_load.p0.v[i] * sbase,
                       Q=ssa_load.q0.v[i] * sbase)

        ssp.add_load(dic_bus[ssa_load.bus.v[i]], load)

    return ssp


def _to_gc_shunt(ssp, ssa_shunt, dic_bus, sbase=1.0):
    """
    Define shunts in GridCal's grid.
    """

    for i in range(len(ssa_shunt)):
        shunt = gc.Shunt(name=ssa_shunt.name.v[i],
                         active=ssa_shunt.u.v[i],
                         G=ssa_shunt.g.v[i] * sbase,
                         B=ssa_shunt.b.v[i] * sbase)

        ssp.add_shunt(dic_bus[ssa_shunt.bus.v[i]], shunt)

    return ssp


def _to_gc_generator(ssp, ssa_slack, ssa_pv, dic_bus, sbase=1.0):
    """
    Define generators considering slack and PV buses.
    """

    for i in range(len(ssa_slack)):

        sl_name = dic_bus[ssa_slack.bus.v[i]]
        b2_dict = ssp.get_bus_names()
        bus_id = b2_dict.index(sl_name.name)
        ssp.buses[bus_id].is_slack = True

        gen = gc.Generator(name=str(ssa_slack.SynGen.v[i]),
                           active=ssa_slack.u.v[i],
                           active_power=ssa_slack.p0.v[i] * sbase,
                           power_factor=ssa_slack.p0.v[i] / np.sqrt((ssa_slack.p0.v[i]**2 + ssa_slack.q0.v[i]**2)),
                           p_min=ssa_slack.pmin.v[i] * sbase,
                           p_max=ssa_slack.pmax.v[i] * sbase,
                           Qmin=ssa_slack.qmin.v[i] * sbase,
                           Qmax=ssa_slack.qmax.v[i] * sbase,
                           voltage_module=ssa_slack.v0.v[i],
                           Snom=ssa_slack.Sn.v[i])

        ssp.add_generator(dic_bus[ssa_slack.bus.v[i]], gen)

    for i in range(len(ssa_pv)):
        gen = gc.Generator(name=str(ssa_pv.SynGen.v[i]),
                           active=ssa_pv.u.v[i],
                           active_power=ssa_pv.p0.v[i] * sbase,
                           power_factor=ssa_pv.p0.v[i] / np.sqrt((ssa_pv.p0.v[i]**2 + ssa_pv.q0.v[i]**2)),
                           p_min=ssa_pv.pmin.v[i] * sbase,
                           p_max=ssa_pv.pmax.v[i] * sbase,
                           Qmin=ssa_pv.qmin.v[i] * sbase,
                           Qmax=ssa_pv.qmax.v[i] * sbase,
                           voltage_module=ssa_pv.v0.v[i],
                           Snom=ssa_pv.Sn.v[i])

        ssp.add_generator(dic_bus[ssa_pv.bus.v[i]], gen)

    return ssp


@require_gridcal
def to_gridcal(ssa, verify=True, tol=1e-6):
    """
    Convert an ANDES system to a GridCal grid.

    Parameters
    ----------
    ssa : andes.system.System
        The ANDES system to be converted
    verify : bool
        If True, the converted network will be verified with the source ANDES
        system using AC power flow.
    tol : float
        The tolerance of error when comparing power flow solutions.

    Returns
    -------
    GridCal.Engine.Core.multi_circuit.MultiCircuit
        A GridCal net with the same bus, branch, gen, and load data as the ANDES
        system

    Notes
    -----
    Handling of the following parameters:

      - By default, all generators in ``ssp`` are controllable unless
        user-defined controllability is given
      - The online status of generators are determined by the online status of
        ``StaticGen`` that connected to the ``SynGen`` or ``DG``
      - ``ssp.gen.name`` is from ``ssa.StaticGen.idx``, which should be unique

    """

    # create an empty GC grid
    sbase = ssa.config.mva
    ssp = gc.MultiCircuit(Sbase=sbase, fbase=ssa.config.freq, name=ssa.name)

    # 1. convert buses
    ssp, dic_bus = _to_gc_bus(ssp, ssa.Bus)

    # 2. convert branches
    ssp = _to_gc_branch(ssp, ssa.Line, dic_bus)

    # 3. convert loads
    ssp = _to_gc_load(ssp, ssa.PQ, dic_bus, sbase=sbase)

    # 4. convert shunts
    ssp = _to_gc_shunt(ssp, ssa.Shunt, dic_bus, sbase=sbase)

    # 5. convert generators (Slack and PV)
    ssp = _to_gc_generator(ssp, ssa.Slack, ssa.PV, dic_bus, sbase=sbase)

    if verify:
        _verify_pf(ssa, ssp, tol)

    return ssp


def _verify_pf(ssa, ssp, tol=1e-6):
    """
    Verify power flow results.
    """

    # ANDES
    ssa.PFlow.run()
    pf_bus = ssa.Bus.as_df()[["name"]]

    pf_bus['v_andes'] = ssa.Bus.v.v
    pf_bus['a_andes'] = ssa.Bus.a.v

    # GridCal
    options = gc.PowerFlowOptions(gc.SolverType.NR, verbose=False)
    pf = gc.PowerFlowDriver(ssp, options)
    pf.run()

    # GridCal assumes the slack has an angle of 0 always, correct it
    pf.results.voltage = pf.results.voltage * np.exp(1j * ssa.Slack.a0.v[0])

    # Check
    vm_dif = pf_bus['v_andes'] - np.abs(pf.results.voltage)
    va_dif = pf_bus['a_andes'] - np.angle(pf.results.voltage)

    ret = False
    if (np.max(np.abs(vm_dif)) < tol) and (np.max(np.abs(va_dif)) < tol):
        logger.info("Power flow results are consistent. Conversion is successful.")
        ret = True
    else:
        logger.warning("Warning: Power flow results are inconsistent. Please check!")
        logger.warning(vm_dif)
        logger.warning(va_dif)

    return ret



"""
HYGOV4 hydro governor model.
"""

from andes.core import Algeb, ConstService, NumParam
from andes.core.block import Integrator, IntegratorAntiWindup, Lag, Washout, GainLimiter
from andes.models.governor.tgbase import TGBase, TGBaseData


class HYGOV4Data(TGBaseData):
    """
    HYGOV4 Data
    """

    def __init__(self):
        super().__init__()
        self.Rperm = NumParam(info='Speed Regulation Gain (mach. base default)',
                              tex_name='R_{perm}',
                              default=0.5,
                              unit='p.u.',
                              ipower=True,
                              )
        self.Rtemp = NumParam(info='Temporary Droop (Rtemp < Rperm)',
                              tex_name='R_{temp}',
                              default=1,
                              unit='p.u.',
                              ipower=True,
                              )
        self.UO = NumParam(info='Maximum Gate opening velocity',
                           tex_name='U_O',
                           default=1,
                           unit='p.u.',
                           power=True,
                           )
        self.UC = NumParam(info='Maximum Gate closing velocity',
                           tex_name='U_C',
                           default=0,
                           unit='p.u.',
                           power=True,
                           )
        self.PMAX = NumParam(info='Maximum Gate opening',
                             tex_name='P_{MAX}',
                             default=1,
                             unit='p.u.',
                             power=True,
                             )
        self.PMIN = NumParam(info='Minimum Gate opening',
                             tex_name='P_{MIN}',
                             default=0,
                             unit='p.u.',
                             power=True,
                             )
        self.Tp = NumParam(info='Pilot servo time constant',
                           default=0.05,
                           tex_name='T_p'
                           )
        self.Tg = NumParam(info='Gate servo time constant',
                           default=0.05,
                           tex_name='T_g'
                           )
        self.Tr = NumParam(info='Dashpot time constant',
                           default=0.05,
                           tex_name='T_r'
                           )
        self.Tw = NumParam(info='Water inertia time constant',
                           default=1,
                           tex_name='T_w'
                           )
        self.At = NumParam(info='Turbine gain',
                           default=1,
                           tex_name='A_t'
                           )
        self.Dturb = NumParam(info='Turbine Damping Factor',
                              default=0.0,
                              tex_name='D_{turb}',
                              power=True,
                              )
        self.Hdam = NumParam(info='Head available at dam',
                             default=1,
                             tex_name='H_{dam}',
                             power=True,
                             )
        self.qNL = NumParam(info='No-Load flow at nominal head',
                            default=0.1,
                            tex_name='q_{NL}',
                            power=True,
                            )

        # nonlinear points, keep for future use
        # self.GV0 = NumParam(info='Governor point 0',
        #                     default=0.1,
        #                     tex_name='GV_0')
        # self.PGV0 = NumParam(info='Governor power point 0',
        #                      default=0,
        #                      tex_name='PGV_0')
        # self.GV1 = NumParam(info='Governor point 1',
        #                     default=0.15,
        #                     tex_name='GV_1')
        # self.PGV1 = NumParam(info='Governor power point 1',
        #                      default=0.1,
        #                      tex_name='PGV_1')
        # self.GV2 = NumParam(info='Governor point 2',
        #                     default=0.3,
        #                     tex_name='GV_2')
        # self.PGV2 = NumParam(info='Governor power point 2',
        #                      default=0.5,
        #                      tex_name='PGV_2')
        # self.GV3 = NumParam(info='Governor point 3',
        #                     default=0.5,
        #                     tex_name='GV_3')
        # self.PGV3 = NumParam(info='Governor power point 3',
        #                      default=0.7,
        #                      tex_name='PGV_3')
        # self.GV4 = NumParam(info='Governor point 4',
        #                     default=0.6,
        #                     tex_name='GV_4')
        # self.PGV4 = NumParam(info='Governor power point 4',
        #                      default=0.8,
        #                      tex_name='PGV_4')
        # self.GV5 = NumParam(info='Governor point 5',
        #                     default=0.7,
        #                     tex_name='GV_5')
        # self.PGV5 = NumParam(info='Governor power point 5',
        #                      default=0.9,
        #                      tex_name='PGV_5')


class HYGOV4Model(TGBase):
    """
    Implement HYGOV4 model.
    """

    def __init__(self, system, config):
        TGBase.__init__(self, system, config)

        self.iTg = ConstService(v_str='u/Tg',
                                tex_name='1/T_g',
                                )
        self.R = ConstService(v_str='Rtemp + Rperm',
                              tex_name='R_{temp} + R_{perm}',
                              )
        self.TrRtemp = ConstService(v_str='Rtemp * Tr',
                                    tex_name='R_{temp} * T_r',
                                    )
        self.q0 = ConstService(v_str='tm0 / (At * Hdam) + qNL',
                               tex_name='q_0',
                               )
        self.pref = Algeb(info='Reference power input',
                          tex_name='P_{ref}',
                          v_str='Rperm * (q0 / (Hdam ** 0.5))',
                          e_str='Rperm * (q0 / (Hdam ** 0.5)) - pref'
                          )

        self.wd = Algeb(info='Generator speed deviation',
                        unit='p.u.',
                        tex_name=r'\omega_{dev}',
                        v_str='0',
                        e_str='ue * (omega - wref) - wd',
                        )

        self.SV = GainLimiter(u='LAG_y', K=self.iTg, R=1,
                              upper=self.UO, lower=self.UC,
                              info='servo gain and limiters'
                              )

        self.GATE = IntegratorAntiWindup(u='SV_y', upper=self.PMAX, lower=self.PMIN,
                                         T=1, K=1,
                                         y0='(q0 / (Hdam ** 0.5))',
                                         info="Gate position"
                                         )
        self.WO = Washout(u='GATE_y',
                          K=self.TrRtemp,
                          T=self.Tr,
                          info='Washout feedback with T_r'
                          )

        self.Psum = Algeb(info='summation of power input to servo',
                          unit='p.u.',
                          tex_name='P_{sum}',
                          v_str='0',
                          e_str='ue * (pref + paux - (Rperm * GATE_y + WO_y ) - wd) - Psum',
                          )

        self.LAG = Lag(u=self.Psum,
                       K=1,
                       T=self.Tp,
                       info='lag block with T_p, outputs velocity',
                       )

        self.trhead = Algeb(info='turbine head',
                            unit='p.u.',
                            tex_name="trhead",
                            e_str='q_y**2 / GATE_y**2 - trhead',
                            v_str='Hdam',
                            )

        self.q = Integrator(u='Hdam - trhead',
                            T=self.Tw, K=1,
                            y0='q0',
                            check_init=False,
                            info="turbine flow (q)"
                            )

        self.pout.e_str = 'ue * (At * trhead * (q_y - qNL) - Dturb * wd * GATE_y) - pout'


class HYGOV4(HYGOV4Data, HYGOV4Model):
    """
    HYGOV4 turbine governor model.

    Implements the PSS/E HYGOV4 model with the following ignored:

    - input deadband DB1
    - valve position deadband DB2
    - nonlinear function bewteen GV and P_{GV}
    """

    def __init__(self, system, config):
        HYGOV4Data.__init__(self)
        HYGOV4Model.__init__(self, system, config)



"""
Module for renewable energy generator (converter) model A.
"""

from andes.core import (Algeb, ConstService, ExtAlgeb, ExtParam, ExtService,
                        IdxParam, Lag, Model, ModelData, NumParam, Piecewise,)
from andes.core.block import GainLimiter, LagAntiWindupRate
from andes.core.service import PostInitService
from andes.core.var import AliasAlgeb


class REGCA1Data(ModelData):
    """
    REGC_A model data.

    12/09/2020: Made `Lvplsw` a parameter.
    """

    def __init__(self):
        ModelData.__init__(self)

        self.bus = IdxParam(model='Bus',
                            info="interface bus id",
                            mandatory=True,
                            )
        self.gen = IdxParam(info="static generator index",
                            mandatory=True,
                            )
        self.Sn = NumParam(default=100.0, tex_name='S_n',
                           info='Model MVA base',
                           unit='MVA',
                           )

        self.Tg = NumParam(default=0.1, tex_name='T_g',
                           info='converter time const.', unit='s',
                           )
        self.Rrpwr = NumParam(default=10.0, tex_name='R_{rpwr}',
                              info='Low voltage power logic (LVPL) ramp limit',
                              unit='p.u.',
                              )
        self.Brkpt = NumParam(default=1.0, tex_name='B_{rkpt}',
                              info='LVPL characteristic voltage 2',
                              unit='p.u.',
                              )
        self.Zerox = NumParam(default=0.5, tex_name='Z_{erox}',
                              info='LVPL characteristic voltage 1',
                              unit='p.u',
                              )
        # TODO: ensure Brkpt > Zerox
        self.Lvplsw = NumParam(default=1.0, tex_name='z_{Lvplsw}',
                               info='Low volt. P logic: 1-enable, 0-disable',
                               unit='bool',
                               )

        self.Lvpl1 = NumParam(default=1.0, tex_name='L_{vpl1}',
                              info='LVPL gain at Brkpt',
                              unit='p.u',
                              )
        self.Volim = NumParam(default=1.2, tex_name='V_{olim}',
                              info='Voltage lim for high volt. reactive current mgnt.',
                              unit='p.u.',
                              )
        self.Lvpnt1 = NumParam(default=0.8, tex_name='L_{vpnt1}',
                               info='High volt. point for low volt. active current mgnt.',
                               unit='p.u.',
                               )
        self.Lvpnt0 = NumParam(default=0.4, tex_name='L_{vpnt0}',
                               info='Low volt. point for low volt. active current mgnt.',
                               unit='p.u.',
                               )
        # TODO: ensure Lvpnt1 > Lvpnt0
        self.Iolim = NumParam(default=-1.5, tex_name='I_{olim}',
                              info='lower current limit for high volt. reactive current mgnt.',
                              unit='p.u. (mach base)',
                              current=True,
                              )
        self.Tfltr = NumParam(default=0.1, tex_name='T_{fltr}',
                              info='Voltage filter T const for low volt. active current mgnt.',
                              unit='s',
                              )
        self.Khv = NumParam(default=0.7, tex_name='K_{hv}',
                            info='Overvolt. compensation gain in high volt. reactive current mgnt.',
                            )
        self.Iqrmax = NumParam(default=1, tex_name='I_{qrmax}',
                               info='Upper limit on the ROC for reactive current',
                               unit='p.u.',
                               current=True,
                               )
        self.Iqrmin = NumParam(default=-1, tex_name='I_{qrmin}',
                               info='Lower limit on the ROC for reactive current',
                               unit='p.u.',
                               current=True,
                               )
        self.Accel = NumParam(default=0.0, tex_name='A_{ccel}',
                              info='Acceleration factor',
                              vrange=(0, 1.0),
                              )
        self.gammap = NumParam(default=1.0,
                               info="P ratio of linked static gen",
                               tex_name=r'\gamma_P'
                               )
        self.gammaq = NumParam(default=1.0,
                               info="Q ratio of linked static gen",
                               tex_name=r'\gamma_Q'
                               )


class REGCA1Model(Model):
    """
    REGCA1 implementation.
    """

    def __init__(self, system, config):
        Model.__init__(self, system, config)
        self.flags.tds = True
        self.group = 'RenGen'

        self.a = ExtAlgeb(model='Bus',
                          src='a',
                          indexer=self.bus,
                          tex_name=r'\theta',
                          info='Bus voltage angle',
                          e_str='-Pe',
                          ename='P',
                          tex_ename='P',
                          )

        self.v = ExtAlgeb(model='Bus',
                          src='v',
                          indexer=self.bus,
                          tex_name=r'V',
                          info='Bus voltage magnitude',
                          e_str='-Qe',
                          ename='Q',
                          tex_ename='Q',
                          )

        self.vd = AliasAlgeb(self.v)

        self.p0s = ExtService(model='StaticGen',
                              src='p',
                              indexer=self.gen,
                              tex_name='P_{0s}',
                              info='initial P of the static gen',
                              )
        self.q0s = ExtService(model='StaticGen',
                              src='q',
                              indexer=self.gen,
                              tex_name='Q_{0s}',
                              info='initial Q of the static gen',
                              )
        self.p0 = ConstService(v_str='p0s * gammap',
                               tex_name='P_0',
                               info='initial P of this gen',
                               )
        self.q0 = ConstService(v_str='q0s * gammaq',
                               tex_name='Q_0',
                               info='initial Q of this gen',
                               )
        self.ra = ExtParam(model='StaticGen',
                           src='ra',
                           indexer=self.gen,
                           tex_name='r_a',
                           export=False,
                           )
        self.xs = ExtParam(model='StaticGen',
                           src='xs',
                           indexer=self.gen,
                           tex_name='x_s',
                           export=False,
                           )

        # --- INITIALIZATION ---
        self.q0gt0 = ConstService('Indicator(q0> 0)', tex_name='z_{q0>0}',
                                  info='flags for q0 below zero',
                                  )
        self.q0lt0 = ConstService('Indicator(q0< 0)', tex_name='z_{q0<0}',
                                  info='flags for q0 below zero',
                                  )

        self.Ipcmd0 = ConstService('p0 / v', info='initial Ipcmd',
                                   tex_name='I_{pcmd0}',
                                   )

        self.Iqcmd0 = ConstService('-q0 / v', info='initial Iqcmd',
                                   tex_name='I_{qcmd0}',
                                   )

        self.LVG = Piecewise(u=self.v, points=('Lvpnt0', 'Lvpnt1'),
                             funs=('0', '(v - Lvpnt0) * kLVG', '1'),
                             info='Ip gain during low voltage',
                             tex_name='L_{VG}',
                             )

        # `Ipcmd` is not defined when the initial `LVG_y` is zero
        self.Ipcmd = Algeb(tex_name='I_{pcmd}',
                           info='current component for active power',
                           e_str='Ipcmd0_LVG - Ipcmd',
                           v_str='Indicator(LVG_y>0) * Ipcmd0 / LVG_y + Indicator(LVG_y<=0) * 1',
                           diag_eps=True,
                           )

        self.Ipcmd0_LVG = PostInitService(v_str='Ipcmd',
                                          info='initial Ipcmd considering initial LVG output')

        self.Iqcmd = Algeb(tex_name='I_{qcmd}',
                           info='current component for reactive power',
                           e_str='Iqcmd0 - Iqcmd', v_str='Iqcmd0')

        # reactive power management

        # rate limiting logic (for fault recovery, although it does not detect any recovery)
        #   - activate upper limit when q0 > 0 (self.q0gt0)
        #   - activate lower limit when q0 < 0 (self.q0lt0)

        self.S1 = LagAntiWindupRate(u=self.Iqcmd,
                                    T=self.Tg, K=-1,
                                    lower=-9999, upper=9999, no_lower=True, no_upper=True,
                                    rate_lower=self.Iqrmin, rate_upper=self.Iqrmax,
                                    rate_lower_cond=self.q0lt0, rate_upper_cond=self.q0gt0,
                                    tex_name='S_1',
                                    info='Iqcmd delay',
                                    )  # output `S1_y` == `Iq`

        # piece-wise gain for low voltage active current mgnt.
        self.kLVG = ConstService(v_str='1 / (Lvpnt1 - Lvpnt0)',
                                 tex_name='k_{LVG}',
                                 )

        # piece-wise gain for LVPL
        self.kLVPL = ConstService(v_str='Lvplsw * Lvpl1 / (Brkpt - Zerox)',
                                  tex_name='k_{LVPL}',
                                  )

        self.S2 = Lag(u=self.v, T=self.Tfltr, K=1.0,
                      info='Voltage filter with no anti-windup',
                      tex_name='S_2',
                      )
        self.LVPL = Piecewise(u=self.S2_y,
                              points=('Zerox', 'Brkpt'),
                              funs=('0 + 9999*(1-Lvplsw)',
                                    '(S2_y - Zerox) * kLVPL + 9999 * (1-Lvplsw)',
                                    '9999'),
                              info='Low voltage Ipcmd upper limit',
                              tex_name='L_{VPL}',
                              )

        self.S0 = LagAntiWindupRate(u=self.Ipcmd, T=self.Tg, K=1,
                                    upper=self.LVPL_y, rate_upper=self.Rrpwr,
                                    lower=-9999, rate_lower=-9999,
                                    no_lower=True, rate_no_lower=True,
                                    tex_name='S_0',
                                    )  # `S0_y` is the output `Ip` in the block diagram

        self.Ipout = Algeb(e_str='S0_y * LVG_y -Ipout',
                           v_str='Ipcmd * LVG_y',
                           info='Output Ip current',
                           tex_name='I_{pout}',
                           )

        # high voltage part
        self.HVG = GainLimiter(u='v - Volim',
                               K=self.Khv, R=1,
                               info='High voltage gain block',
                               lower=0, upper=999, no_upper=True,
                               tex_name='H_{VG}'
                               )
        self.HVG.lim.no_warn = True
        self.HVG.lim.allow_adjust = False

        self.Iqout = GainLimiter(u='S1_y - HVG_y',
                                 K=1, R=1,
                                 lower=self.Iolim, upper=9999,
                                 no_upper=True, info='Iq output block',
                                 tex_name='I^{qout}',
                                 )  # `Iqout_y` is the final Iq output

        self.Pe = Algeb(tex_name='P_e', info='Active power output',
                        v_str='p0', e_str='Ipout * v - Pe')
        self.Qe = Algeb(tex_name='Q_e', info='Reactive power output',
                        v_str='q0', e_str='Iqout_y * v - Qe')

    def v_numeric(self, **kwargs):
        """
        Disable the corresponding `StaticGen`s.
        """
        self.system.groups['StaticGen'].set(src='u', idx=self.gen.v, attr='v', value=0)


class REGCA1(REGCA1Data, REGCA1Model):
    """
    Renewable energy generator model type A.

    Implements ``REGCA1`` in PSS/E, or ``REGC_A`` in PSLF and Powerworld.

    Volim is the voltage limit for high voltage reactive current management,
    which should be large than static bus voltage (Volim > v),
    or initialization error will occur.
    """

    def __init__(self, system, config):
        REGCA1Data.__init__(self)
        REGCA1Model.__init__(self, system, config)



"""
EXAC2 exciter.
"""

from andes.core.block import GainLimiter, LVGate
from andes.core.param import NumParam
from andes.core.service import PostInitService
from andes.core.var import Algeb
from andes.models.exciter.exac1 import EXAC1Data, EXAC1Model


class EXAC2Data(EXAC1Data):
    """
    EXAC2 parameters.
    """

    def __init__(self):
        EXAC1Data.__init__(self)
        self.VAMAX = NumParam(info='Maximum KA block output',
                              tex_name='V_{RMAX}',
                              default=8,
                              unit='p.u.',
                              vrange=(0.5, 10),
                              )
        self.VAMIN = NumParam(info='Minimum KA block output',
                              tex_name='V_{RMIN}',
                              default=0,
                              unit='p.u.',
                              vrange=(-10, 0.5),
                              )
        self.VLR = NumParam(info='low voltage constant',
                            tex_name='V_{LR}',
                            default=0.0,
                            unit='p.u.',
                            )
        self.KL = NumParam(info='gain for low voltage',
                           tex_name='K_L',
                           default=1.0,
                           unit='p.u.',
                           )
        self.KH = NumParam(info='gain for high voltage',
                           tex_name='K_H',
                           default=1.0,
                           unit='p.u.',
                           )
        self.KB = NumParam(info='gain KB for regulator',
                           tex_name='K_B',
                           default=1.0,
                           unit='p.u.',
                           non_negative=True,
                           )


class EXAC2Model(EXAC1Model):
    """
    EXAC2 implementation.
    """

    def __init__(self, system, config):
        EXAC1Model.__init__(self, system, config)

        self.VHA = Algeb(v_str='LA_y - KH * VFE',
                         e_str='LA_y - KH * VFE - VHA',
                         )
        self.VL = Algeb(v_str='(VLRx - VFE) * KL',
                        e_str='(VLRx - VFE) * KL - VL')

        self.LA.lower = self.VAMIN
        self.LA.upper = self.VAMAX

        self.LVG = LVGate(u1=self.VHA, u2=self.VL)

        self.VR = GainLimiter(u='LVG_y', K=self.KB, R=1,
                              upper=self.VRMAX, lower=self.VRMIN,
                              )

        self.INT.u = 'ue* (VR_y - VFE)'  # this won't propagate to its `y` variable. Need the next line
        self.INT_y.e_str = 'ue* (VR_y - VFE)'

        self.vref.v_str = '(VFE * KL + VFE / KB) / KA + v'

        self.VLRx = Algeb(v_str='Indicator(VFE / KL / KB + VFE > VLR) * (VFE / KL / KB + VFE) + '
                          'Indicator(VFE / KL / KB + VFE < VLR) * VLR ',
                          e_str='VLR0 - VLRx'
                          )

        self.VLR0 = PostInitService(v_str='VLRx')


class EXAC2(EXAC1Data, EXAC1Model):
    """
    EXAC2 model.

    Ref: https://www.powerworld.com/WebHelp/Content/TransientModels_HTML/Exciter%20EXAC2.htm

    Notes
    -----

    ``VLR`` is an input parameter, but to initialize the LVGate, an internal
    ``VLRx`` will be computed as a contant upon initialization. The constant
    ``VLRx`` will be used in the place of ``VLR`` in the block diagram.
    """

    def __init__(self, system, config):
        EXAC2Data.__init__(self)
        EXAC2Model.__init__(self, system, config)



"""
Tests for configurations.
"""

import unittest

import andes


class TestConfigOption(unittest.TestCase):
    """
    Tests for config_option.
    """

    def test_config_option(self):
        """
        Test single and multiple config_option passed to `andes.run`.
        """

        path = andes.get_case("5bus/pjm5bus.json")
        self.assertRaises(ValueError, andes.load, path, config_option={"TDS = 1"})
        self.assertRaises(ValueError, andes.load, path, config_option={"System.TDS.any = 1"})
        self.assertRaises(ValueError, andes.load, path, config_option={"TDS.tf == 1"})

        ss = andes.load(path, config_option={"PQ.pq2z = 0"}, default_config=True)
        self.assertEqual(ss.PQ.config.pq2z, 0)

        ss = andes.load(path, config_option={"PQ.pq2z=0"}, default_config=True)
        self.assertEqual(ss.PQ.config.pq2z, 0)

        ss = andes.load(path, config_option=["PQ.pq2z=0", "TDS.tf = 1"], default_config=True)
        self.assertEqual(ss.PQ.config.pq2z, 0)
        self.assertEqual(ss.TDS.config.tf, 1)



"""
Scipy sparse linear solver with SuperLU backend.
"""

import numpy as np

from scipy.sparse import csc_matrix
from scipy.sparse.linalg import spsolve, splu


class SciPySolver:
    """
    Base class for scipy family solvers.
    """

    def __init__(self):
        self.factorize = True

        # when `new_A` is True, rebuild and factorize A
        self.new_A = True

    def solve(self, A, b):
        """
        Solve linear systems.

        Parameters
        ----------
        A : scipy.csc_matrix
            Sparse N-by-N matrix
        b : numpy.ndarray
            Dense 1-dimensional array of size N

        Returns
        -------
        np.ndarray
            Solution x to `Ax = b`

        """
        raise NotImplementedError

    def linsolve(self, A, b):
        """
        Exactly same functionality as `solve`.
        """
        return self.solve(A, b)

    def clear(self):
        pass


class SpSolve(SciPySolver):
    """
    scipy.sparse.linalg.spsolve Solver.
    """

    def solve(self, A, b):

        if self.factorize or self.new_A:
            A_csc = spmatrix_to_csc(A)
            self.lu = splu(A_csc)

            self.factorize = False
            self.new_A = False

        x = self.lu.solve(np.ravel(b))
        return x

    def linsolve(self, A, b):
        """
        Solve using `spsolve`.
        """

        A_csc = spmatrix_to_csc(A)
        b = np.ravel(b)
        return spsolve(A_csc, b)


def spmatrix_to_csc(A):
    """
    Convert A of ``kvxopt.spmatrix`` to ``scipy.sparse.csc_matrix``.

    Parameters
    ----------
    A : kvxopt.spmatrix
        Sparse N-by-N matrix

    Returns
    -------
    scipy.sparse.csc_matrix
        Converted csc_matrix

    """
    ccs = A.CCS
    size = A.size
    data = np.array(ccs[2]).ravel()
    indices = np.array(ccs[1]).ravel()
    indptr = np.array(ccs[0]).ravel()
    return csc_matrix((data, indices, indptr), shape=size)



"""
Grid-forming renewable energy model with VSM control.
"""

from andes.core.var import Algeb, ExtAlgeb
from andes.core.service import DeviceFinder, ConstService
from andes.core.param import IdxParam, NumParam
from andes.core.block import Integrator, GainLimiter
from andes.models.renewable.regf1 import REGF1Data, REGF1Model, REGFInnerPIModel, REGFOuterPIModel


class REGF2Data(REGF1Data):
    """
    REGF2 model data.
    """

    def __init__(self):
        REGF1Data.__init__(self)

        self.mf = NumParam(default=0.15, info='VSM inertia constant',
                           tex_name='M_f'
                           )
        self.dd = NumParam(default=0.11, info='VSM damping factor',
                           tex_name='d_d')
        self.pll = IdxParam(info='PLL device idx (optional)', default=None, model='PLL')


class REGF2Primary:
    """
    Primary frequency and voltage controllers based on VSM.
    """

    def __init__(self) -> None:

        self.pllidx = DeviceFinder(self.pll,
                                   link=self.bus,
                                   idx_name='bus',
                                   default_model='PLL2',
                                   auto_find=True,
                                   auto_add=True,
                                   )

        self.plldw = ExtAlgeb(model='PLL',
                              src='PI_y',
                              indexer=self.pllidx,
                              allow_none=True,
                              tex_name=r'\Delta \omega_{PLL}',
                              info='PLL measured freq. deviation',
                              )
        self.wref = Algeb(v_str='1', e_str='1 - wref',
                          tex_name=r'\omega_{ref}', info='speed ref', unit='pu')

        self.Tint = ConstService(v_str='mf * wdrp', tex_name='T_{int}')

        self.INTw = Integrator(u='(plldw * dd * wdrp) + wref + wdrp * (PIplim_y - Psen_y) - INTw_y',
                               T=self.Tint, K=1, y0='wref',
                               )

        self.dw = GainLimiter(u='w0 * (INTw_y - wref)', K=1, R=1,
                              lower=self.dwmin,
                              upper=self.dwmax,
                              )

        self.vref2 = Algeb(tex_name=r'v_{ref2}',
                           info='voltage reference after droop',
                           e_str='(u * PIqlim_y - Qsen_y) * Qdrp + vref - vref2',
                           v_str='u * vref')


class REGF2(REGF2Data, REGF1Model, REGF2Primary,
            REGFOuterPIModel, REGFInnerPIModel):
    """
    Grid-forming inverter with VSM control.

    Implementation of EPRI Memorandum

    D. Ramasubramanian, "PROPOSAL FOR SUITE OF GENERIC GRID FORMING (GFM) POSITIVE SEQUENCE MODELS"
    """

    def __init__(self, system, config):
        REGF2Data.__init__(self)

        REGF1Model.__init__(self, system, config)
        REGF2Primary.__init__(self)
        REGFOuterPIModel.__init__(self)
        REGFInnerPIModel.__init__(self)



"""
Module for Fortescue's symmetric component calculation.
"""

from andes.core.model.modeldata import ModelData
from andes.core.model.model import Model
from andes.core.param import IdxParam, NumParam
from andes.core.service import ConstService
from andes.core.var import Algeb, ExtAlgeb


class FortescueData(ModelData):
    """
    Fortescue data.
    """

    def __init__(self):
        super().__init__()

        self.bus = IdxParam(model='Bus', info="bus idx for the single-phase equivalent", mandatory=True)
        self.busa = IdxParam(model='Bus', info="bus idx for phase a", mandatory=True)
        self.busb = IdxParam(model='Bus', info="bus idx for phase b", mandatory=True)
        self.busc = IdxParam(model='Bus', info="bus idx for phase c", mandatory=True)

        self.Sn = NumParam(default=100.0,
                           info="Power rating",
                           non_zero=True,
                           tex_name=r'S_n',
                           unit='MW',
                           )

        self.r = NumParam(default=1e-3,
                          info="resistance",
                          tex_name='r',
                          z=True,
                          unit='p.u.',
                          )
        self.x = NumParam(default=0.025,
                          info="short-circuit reactance",
                          tex_name='x',
                          z=True,
                          unit='p.u.',
                          non_zero=True,
                          )
        self.g = NumParam(default=0.0,
                          info="iron loss",
                          y=True,
                          unit='p.u.',
                          )
        self.b = NumParam(default=0.005,
                          info="magnetizing susceptance",
                          y=True,
                          unit='p.u.',
                          )


class FortescueModel(Model):
    """
    Model implementation.
    """

    def __init__(self, system=None, config=None):
        super().__init__(system, config)

        self.group = 'Interface'
        self.flags.pflow = True
        self.flags.tds = True

        self.a = ExtAlgeb(model='Bus', src='a', indexer=self.bus, tex_name=r'\theta_1',
                          info='phase angle of single-phase eq. bus',
                          ename='P1',
                          tex_ename='P_1',
                          )
        self.aa = ExtAlgeb(model='Bus', src='a', indexer=self.busa, tex_name=r'\theta_a',
                           info='phase angle of bus for phase a',
                           ename='Pa',
                           tex_ename='P_{a}',
                           diag_eps=True,
                           )
        self.ab = ExtAlgeb(model='Bus', src='a', indexer=self.busb, tex_name=r'\theta_b',
                           info='phase angle of bus for phase b',
                           ename='Pb',
                           tex_ename='P_{b}',
                           diag_eps=True,
                           )
        self.ac = ExtAlgeb(model='Bus', src='a', indexer=self.busc, tex_name=r'\theta_c',
                           info='phase angle of bus for phase c',
                           ename='Pc',
                           tex_ename='P_{c}',
                           diag_eps=True,
                           )

        self.v = ExtAlgeb(model='Bus', src='v', indexer=self.bus, tex_name='V_1',
                          info='voltage of single-phase eq. bus',
                          ename='Q1',
                          tex_ename='Q_1',
                          )
        self.va = ExtAlgeb(model='Bus', src='v', indexer=self.busa, tex_name='V_a',
                           info='voltage of bus for phase a',
                           ename='Qa',
                           tex_ename='Q_{a}',
                           diag_eps=True,
                           )
        self.vb = ExtAlgeb(model='Bus', src='v', indexer=self.busb, tex_name='V_b',
                           info='voltage of bus for phase b',
                           ename='Qb',
                           tex_ename='Q_{b}',
                           diag_eps=True,
                           )
        self.vc = ExtAlgeb(model='Bus', src='v', indexer=self.busc, tex_name='V_c',
                           info='voltage of bus for phase c',
                           ename='Qc',
                           tex_ename='Q_{c}',
                           diag_eps=True,
                           )

        # gh = g, bh = b, yh = (gh + 1j bh), yk = 0, yhk = u/(r+1j*x)
        self.yhk = ConstService(tex_name='y_{hk}', vtype=complex)
        self.ghk = ConstService(tex_name='g_{hk}')
        self.bhk = ConstService(tex_name='b_{hk}')

        self.yhk.v_str = 'u/(r + 1j*x)'
        self.ghk.v_str = 're(yhk)'
        self.bhk.v_str = 'im(yhk)'

        self.d120 = ConstService(v_str='2/3*pi', tex_name='120^o', info='120 degrees')

        # --- positive sequence: magnitude and angle
        self.vp = Algeb(v_str='(va + vb + vc) / 3',
                        e_str='sqrt((vb * cos(ab + d120 - aa) + \
                                     vc * cos(ac - d120 - aa) + va) ** 2 + \
                                    (vb * sin(ab + d120 - aa) + \
                                     vc * sin(ac - d120 - aa)) ** 2) / 3 - vp',
                        info='positive sequence voltage magnitude',
                        tex_name='V_{p}',
                        )

        self.ap = Algeb(v_str='aa + ab + ac',
                        e_str='atan2((vb * sin(ab + d120 - aa) + \
                                      vc * sin(ac - d120 - aa)), \
                                     (vb * cos(ab + d120 - aa) + \
                                      vc * cos(ac - d120 - aa) + va)) + aa - ap',
                        info='positive sequence voltage phase',
                        tex_name=r'\theta_{p}',
                        )

        # --- negative sequence: d- and q-axis

        self.vnd = Algeb(v_str='0.0',
                         e_str='va * cos(aa) + vb * cos(ab - d120) + vc * cos(ac + d120) - vnd',
                         info='negative sequence voltage on d-axis (cos)',
                         tex_name='V_{nd}',
                         )

        self.vnq = Algeb(v_str='0.0',
                         e_str='va * sin(aa) + vb * sin(ab - d120) + vc * sin(ac + d120) - vnq',
                         info='negative sequence voltage on q-axis (sin)',
                         tex_name='V_{nq}',
                         )

        # --- zero sequence: d- and q-axis voltages
        self.vzd = Algeb(v_str='0.0',
                         e_str='va * cos(aa) + vb * cos(ab) + vc * cos(ac) - vzd',
                         info='zero sequence voltage on d-axis (cos)',
                         tex_name='V_{zd}',
                         )

        self.vzq = Algeb(v_str='0.0',
                         e_str='va * sin(aa) + vb * sin(ab) + vc * sin(ac) - vzq',
                         info='zero sequence voltage on q-axis (sin)',
                         tex_name='V_{zq}',
                         )

        # injection on the primary side

        self.a.e_str = 'u * (v ** 2 * (g + ghk) - v * vp * (ghk * cos(a - ap) + bhk * sin(a - ap)))'

        self.v.e_str = 'u * (-v ** 2 * (b + bhk) - v * vp * (ghk * sin(a - ap) - bhk * cos(a - ap)))'

        # injections on the secondary side

        p_phase = 'u/3*({v_phase}**2 *(g + ghk) - ' \
                  'v*{v_phase}*(ghk * cos(a-({a_phase})) - bhk * sin(a-({a_phase}))))'

        q_phase = 'u/3*(-{v_phase}**2 *(b + bhk) + ' \
                  'v *{v_phase}* (ghk * sin(a - ({a_phase})) + bhk * cos(a - ({a_phase}))))'

        self.aa.e_str = p_phase.format(v_phase='va', a_phase='aa')
        self.ab.e_str = p_phase.format(v_phase='vb', a_phase='ab + d120')
        self.ac.e_str = p_phase.format(v_phase='vc', a_phase='ac - d120')

        self.va.e_str = q_phase.format(v_phase='va', a_phase='aa')
        self.vb.e_str = q_phase.format(v_phase='vb', a_phase='ab + d120')
        self.vc.e_str = q_phase.format(v_phase='vc', a_phase='ac - d120')


class Fortescue(FortescueData, FortescueModel):
    """
    Fortescue's symmetric component interface.

    This model interfaces a positive-sequence, single-phase-equivalent bus with
    three buses representing three phases. It is effectively a transformer with
    one terminal on the primary side and three on the secondary. Only the
    positive sequence component on the secondary winding is used for simulation.

    The positive-sequence voltage magnitude and angle of the secondary winding
    are named ``vp`` and ``ap``.

    The negative and zero sequence variables given in the d- and q-axis due the
    angle being undefined when the voltage is zero. The negative sequence
    voltages are ``vnd`` and ``vnq`` for the d- anx q-axis, respectively.
    Likewise, the zero-sequence voltages are ``vzd`` and ``vzq``.

    """

    def __init__(self, system, config):
        FortescueData.__init__(self)
        FortescueModel.__init__(self, system, config)



"""
PQ load model.
"""

from collections import OrderedDict

from andes.core import (ModelData, IdxParam, NumParam, Model,
                        ExtAlgeb, ExtService, ConstService, Limiter)


class PQData(ModelData):
    def __init__(self):
        super().__init__()
        self.bus = IdxParam(model='Bus',
                            info="linked bus idx",
                            mandatory=True,
                            )

        self.Vn = NumParam(default=110,
                           info="AC voltage rating",
                           unit='kV',
                           non_zero=True,
                           tex_name=r'V_n',
                           )
        self.p0 = NumParam(default=0,
                           info='active power load in system base',
                           power=False,
                           tex_name=r'p_0',
                           unit='p.u.',
                           )
        self.q0 = NumParam(default=0,
                           info='reactive power load in system base',
                           power=False,
                           tex_name=r'q_0',
                           unit='p.u.',
                           )
        self.vmax = NumParam(default=1.2,
                             info='max voltage before switching to impedance',
                             tex_name=r'v_{max}',
                             )
        self.vmin = NumParam(default=0.8,
                             info='min voltage before switching to impedance',
                             tex_name=r'v_{min}',
                             )

        self.owner = IdxParam(model='Owner', info="owner idx")


class PQ(PQData, Model):
    """
    PQ load model.

    Implements an automatic pq2z conversion during power flow when the voltage
    is outside [vmin, vmax]. The conversion can be turned off by setting `pq2z`
    to 0 in the Config file.

    Before time-domain simulation, PQ load will be converted to impedance,
    current source, and power source based on the weights in the Config file.

    Weights (p2p, p2i, p2z) corresponds to the weights for constant power,
    constant current and constant impedance. p2p, p2i and p2z must be in decimal
    numbers and sum up exactly to 1. The same rule applies to (q2q, q2i, q2z).

    To alter the PQ load in terms of power during simulation, one needs to set
    the conversion weights to preserve the constant power portion. For example,
    the PQ can remain as constant power load by setting

    .. code-block :: python

        ss.PQ.config.p2p = 1.0 ss.PQ.config.p2i = 0 ss.PQ.config.p2z = 0

        ss.PQ.config.q2q = 1.0 ss.PQ.config.q2i = 0 ss.PQ.config.q2z = 0

    Then, the constant power portion can be altered by changing the ``Ppf`` and
    ``Qpf`` constants for active power and reactive power.

    The equivalent constant current components are in constants ``Ipeq`` and
    ``Iqeq`` for active and reactive current, and the equivalent impedances are
    in ``Req`` and ``Xeq``.
    """

    def __init__(self, system=None, config=None):
        PQData.__init__(self)
        Model.__init__(self, system, config)
        self.group = 'StaticLoad'
        # ``tds`` flag is needed to retrieve initial voltage (for constant Z/I conversion)
        self.flags.update({'pflow': True,
                           'tds': True,
                           })
        self.config.add(OrderedDict((('pq2z', 1),
                                     ('p2p', 0.0),
                                     ('p2i', 0.0),
                                     ('p2z', 1.0),
                                     ('q2q', 0.0),
                                     ('q2i', 0.0),
                                     ('q2z', 1.0),
                                     )))
        self.config.add_extra("_help",
                              pq2z="pq2z conversion if out of voltage limits",
                              p2p="P constant power percentage for TDS. Must have (p2p+p2i+p2z)=1",
                              p2i="P constant current percentage",
                              p2z="P constant impedance percentage",
                              q2q="Q constant power percentage for TDS. Must have (q2q+q2i+q2z)=1",
                              q2i="Q constant current percentage",
                              q2z="Q constant impedance percentage",
                              )
        self.config.add_extra("_alt",
                              pq2z="(0, 1)",
                              p2p="float",
                              p2i="float",
                              p2z="float",
                              q2q="float",
                              q2i="float",
                              q2z="float",
                              )
        self.config.add_extra("_tex",
                              pq2z="z_{pq2z}",
                              p2p=r"\gamma_{p2p}",
                              p2i=r"\gamma_{p2i}",
                              p2z=r"\gamma_{p2z}",
                              q2q=r"\gamma_{q2q}",
                              q2i=r"\gamma_{q2i}",
                              q2z=r"\gamma_{q2z}",
                              )

        self.a = ExtAlgeb(model='Bus',
                          src='a',
                          indexer=self.bus,
                          tex_name=r'\theta',
                          ename='P',
                          tex_ename='P',
                          is_input=True,
                          )
        self.v = ExtAlgeb(model='Bus',
                          src='v',
                          indexer=self.bus,
                          tex_name=r'V',
                          ename='Q',
                          tex_ename='Q',
                          is_input=True,
                          )

        self.v0 = ExtService(src='v',
                             model='Bus',
                             indexer=self.bus,
                             tex_name='V_0',
                             info='Initial voltage magnitude from power flow'
                             )
        self.a0 = ExtService(src='a',
                             model='Bus',
                             indexer=self.bus,
                             tex_name=r'\theta_0',
                             info='Initial voltage angle from power flow'
                             )

        # Rub, Xub, Rlb, Xlb are constant for both PF and TDS.
        self.Rub = ConstService(info='Equivalent resistance at voltage upper bound',
                                v_str='p0 / vmax**2',
                                tex_name='R_{ub}'
                                )
        self.Xub = ConstService(info='Equivalent reactance at voltage upper bound',
                                v_str='q0 / vmax**2',
                                tex_name='X_{ub}'
                                )
        self.Rlb = ConstService(info='Equivalent resistance at voltage lower bound',
                                v_str='p0 / vmin**2',
                                tex_name='R_{lb}'
                                )
        self.Xlb = ConstService(info='Equivalent reactance at voltage lower bound',
                                v_str='q0 / vmin**2',
                                tex_name='X_{lb}'
                                )

        # Ppf, Qpf, Req, Xeq, Ipeq, Iqeq are only meaningful after initializing TDS
        self.Ppf = ConstService(info='Actual P in power flow',
                                v_str='(p0 * vcmp_zi + Rlb * vcmp_zl * v0**2 + Rub * vcmp_zu * v0**2)',
                                tex_name='P_{pf}')
        self.Qpf = ConstService(info='Actual Q in power flow',
                                v_str='(q0 * vcmp_zi + Xlb * vcmp_zl * v0**2 + Xub * vcmp_zu * v0**2)',
                                tex_name='Q_{pf}')
        self.Req = ConstService(info='Equivalent resistance at steady state',
                                v_str='Ppf / v0**2',
                                tex_name='R_{eq}'
                                )
        self.Xeq = ConstService(info='Equivalent reactance at steady state',
                                v_str='Qpf / v0**2',
                                tex_name='X_{eq}'
                                )
        self.Ipeq = ConstService(info='Equivalent active current source at steady state',
                                 v_str='Ppf / v0',
                                 tex_name='I_{peq}'
                                 )
        self.Iqeq = ConstService(info='Equivalent reactive current source at steady state',
                                 v_str='Qpf / v0',
                                 tex_name='I_{qeq}'
                                 )

        self.vcmp = Limiter(u=self.v,
                            lower=self.vmin,
                            upper=self.vmax,
                            enable=self.config.pq2z,
                            )

        # Note: the "or" condition "|" is not supported in sympy equation strings.
        # They will simply be ignored.

        # To modify P and Q during TDS, use `alter` to set values to `Ppf` and `Qpf`
        # after, before simulation, setting `config.p2p=1` and `config.q2q=1`.

        self.a.e_str = "u * Indicator(dae_t <= 0) * " \
                       "(p0 * vcmp_zi + Rlb * vcmp_zl * v**2 + Rub * vcmp_zu * v**2) + " \
                       "u * Indicator(dae_t > 0) * " \
                       "(p2p * Ppf + p2i * Ipeq * v + p2z * Req * v**2)"

        self.v.e_str = "u * Indicator(dae_t <= 0) * " \
                       "(q0 * vcmp_zi + Xlb * vcmp_zl * v**2 + Xub * vcmp_zu * v**2) + " \
                       "u * Indicator(dae_t > 0) * " \
                       "(q2q * Qpf + q2i * Iqeq * v + q2z * Xeq * v**2)"



import logging
from collections import OrderedDict

import numpy as np

from andes.core.discrete import Switcher
from andes.core.model import Model, ModelData
from andes.core.param import DataParam, IdxParam, NumParam, TimerParam
from andes.core.service import ConstService
from andes.core.var import ExtAlgeb
from andes.shared import tqdm

logger = logging.getLogger(__name__)


class ToggleData(ModelData):
    def __init__(self):
        super(ToggleData, self).__init__()
        self.model = DataParam(info='model or group name of the device',
                               mandatory=True,
                               )
        self.dev = IdxParam(info='idx of the device to control',
                            mandatory=True,
                            )
        self.t = TimerParam(info='switch time for connection status',
                            mandatory=True,
                            )


class Toggle(ToggleData, Model):
    """
    Time-based connectivity status toggle.

    Toggle is used to toggle the connection status (online/offline) of a device
    at the predefined time. Both the model name (or group name) and the device
    idx need to be specified. It effectively negates the ``u`` field of the
    connected device.

    Toggle can be useful to implement disconnection, connection, and
    reconnection of devices. For example, a line trip can be implemented by
    setting ``Line`` to the ``model`` field and the corresponding line's ``idx``
    to the ``dev`` field.

    Multiple Toggles can be added to the same device at different times. Adding
    two Toggles for an initially connected line with ``t=0.1`` and ``t=0.2``,
    for instance, will disconnect the line at t=0.1 sec and reconnect it at
    t=0.2 sec.
    """

    def __init__(self, system, config):
        ToggleData.__init__(self)
        Model.__init__(self, system, config)
        self.flags.update({'tds': True})
        self.group = 'TimedEvent'

        self.t.callback = self._u_switch
        self._init = False   # very first initialization that stores `u`
        self._u = ConstService('1')

    def v_numeric(self, **kwargs):
        """
        Custom initialization function that stores and restores the connectivity status.
        """
        if not self._init:
            for i in range(self.n):
                instance = self.system.__dict__[self.model.v[i]]
                self._u.v[i] = instance.get(src='u', attr='v', idx=self.dev.v[i])
                self._init = True
        else:
            for i in range(self.n):
                instance = self.system.__dict__[self.model.v[i]]
                instance.set(src='u', attr='v', idx=self.dev.v[i], value=self._u.v[i])

    def _u_switch(self, is_time: np.ndarray):
        action = False
        for i in range(self.n):
            if (is_time[i] == 0) or (self.u.v[i] == 0):
                continue

            instance = self.system.__dict__[self.model.v[i]]
            u0 = instance.get(src='u', attr='v', idx=self.dev.v[i])
            instance.set(src='u', attr='v', idx=self.dev.v[i], value=1-u0)
            action = True
            tqdm.write(f'<Toggle {self.idx.v[i]}>: '
                       f'{self.model.v[i]}.{self.dev.v[i]} status '
                       f'changed to {1-u0:g} at t={self.t.v[i]} sec.')
        return action


class Fault(ModelData, Model):
    """
    Three-phase-to-ground fault.

    A Fault device is used to apply and clear three-phase-to-ground fault to the
    given bus. One can set two time parameters, ``tf`` and ``tc``, for the
    fault-on and fault-clearance time, respectively, although only ``tf`` is
    mandatory.

    A fault is implemented by a very small internal shunt impedance to be
    connected at the fault-on time. Its reactance and resistance are specified
    by the parameters ``xf`` and ``rf``.

    To implement a fault and its clearance by tripping a line, one can combine
    ``Fault`` and ``Toggle``. That is, clear a fault in concurrence with a
    Toggle. The user needs to ensure data consistency so that the line trip
    actually clears the fault.

    Non-convergence can occur in the proximity of a fault due to various reasons,
    including network power transfer capability limitation and parameter issues
    of controllers.
    """

    def __init__(self, system, config):
        ModelData.__init__(self)
        self.bus = IdxParam(model='Bus',
                            info="linked bus idx",
                            mandatory=True,
                            )
        self.tf = TimerParam(info='Bus fault start time',
                             unit='second',
                             mandatory=True,
                             callback=self.apply_fault,
                             )
        self.tc = TimerParam(info='Bus fault end time',
                             unit='second',
                             callback=self.clear_fault,
                             )
        self.xf = NumParam(info='Fault to ground reactance (positive)',
                           unit='p.u.(sys)',
                           default=1e-4,
                           tex_name='x_f',
                           )
        self.rf = NumParam(info='Fault to ground resistance (positive)',
                           unit='p.u.(sys)',
                           default=0,
                           tex_name='x_f',
                           )

        Model.__init__(self, system, config)
        self.flags.update({'tds': True})
        self.group = 'TimedEvent'

        self.config.add(OrderedDict((('restore', 1),
                                     ('mode', 1),
                                     ('scale', 1.0),
                                     )))
        self.config.add_extra('_alt',
                              restore=(0, 1),
                              mode=(1, 2, 3),
                              )
        self.config.add_extra('_help',
                              restore='restore algebraic variables to pre-fault values',
                              mode='1. restore all algeb variables, 2. fault bus only',
                              scale='scaling factor of restored algebraic values',
                              )

        self.gf = ConstService(tex_name='g_{f}',
                               v_str='re(1/(rf + 1j * xf))',
                               )
        self.bf = ConstService(tex_name='b_{f}',
                               v_str='im(1/(rf + 1j * xf))',
                               )

        # uf: an internal flag of whether the fault is in action (1) or not (0)
        self.uf = ConstService(tex_name='u_f', v_str='0')

        self.a = ExtAlgeb(model='Bus',
                          src='a',
                          indexer=self.bus,
                          tex_name=r'\theta',
                          info='Bus voltage angle',
                          unit='p.u.(kV)',
                          e_str='u * uf * (v ** 2 * gf)',
                          ename='P',
                          tex_ename='P',
                          )
        self.v = ExtAlgeb(model='Bus',
                          src='v',
                          indexer=self.bus,
                          tex_name=r'V',
                          unit='p.u.(kV)',
                          info='Bus voltage magnitude',
                          e_str='-u * uf * (v ** 2 * bf)',
                          ename='Q',
                          tex_ename='Q',
                          )
        self._vstore = np.array([])

    def apply_fault(self, is_time: np.ndarray):
        """
        Apply fault and store pre-fault algebraic variables (voltages and other
        algebs) to `self._vstore`.
        """
        action = False
        for i in range(self.n):
            if (is_time[i] == 0) or (self.u.v[i] == 0):
                continue

            self.uf.v[i] = 1
            self._vstore = np.array(self.system.dae.y[self.system.Bus.n:])
            logger.debug("Pre-fault algebraic variables:\n" + str(self._vstore))
            tqdm.write(f'<Fault {self.idx.v[i]}>: '
                       f'Applying fault on Bus (idx={self.bus.v[i]}) at t={self.tf.v[i]} sec.')

            action = True
        return action

    def clear_fault(self, is_time: np.ndarray):
        """
        Clear fault and restore pre-fault bus algebraic variables (voltages and
        others).
        """
        action = False
        for i in range(self.n):
            if is_time[i] and (self.u.v[i] == 1):
                self.uf.v[i] = 0

                if self.config.restore:
                    if self.config.mode == 1:
                        self.system.dae.y[self.system.Bus.n:] = self._vstore * self.config.scale
                        logger.debug("All algebraic variables restored after fault clearance at t=%.6f",
                                     self.system.dae.t)

                    # TODO: neither mode 2 or 3 works. Pending further investigation.
                    elif self.config.mode == 2:
                        v_addr = self.system.Bus.get(src='v', idx=self.bus.v[i], attr='a')
                        bus_uid = self.system.Bus.idx2uid(self.bus.v[i])
                        self.system.dae.y[v_addr] = self._vstore[bus_uid] * self.config.scale
                        logger.debug("Voltage on bus %s restored after fault clearance at t=%.6f",
                                     self.bus.v[i], self.system.dae.t)
                    elif self.config.mode == 3:
                        nbus = self.system.Bus.n
                        self.system.dae.y[nbus:2*nbus] = self._vstore[:nbus] * self.config.scale
                        logger.debug("All bus voltages restored after fault clearance at t=%.6f",
                                     self.system.dae.t)
                    else:
                        logger.error("Unsupport fault voltage restoration mode")

                tqdm.write(f'<Fault {self.idx.v[i]}>: '
                           f'Clearing fault on Bus (idx={self.bus.v[i]}) at t={self.tc.v[i]} sec.')

                action = True
        return action


class AlterData(ModelData):
    """
    Data for Alter, which altera values of the given device at a certain time.

    Alter can be used in various timed applications, such as applying load changing,
    tap changing, step response, etc.
    """

    def __init__(self):
        ModelData.__init__(self)
        self.t = TimerParam(info='switch time for connection status', mandatory=True)

        self.model = DataParam(info='model or group name of the device', mandatory=True)
        self.dev = IdxParam(info='idx of the device to alter', mandatory=True)
        self.src = IdxParam(info='model source field (param or service)', mandatory=True)
        self.attr = IdxParam(info='attribute (e.g., v) of the source field', default='v')

        self.method = NumParam(info='alteration method in `+`, `-`, `*`, `/`, `=`',
                               mandatory=True, vtype=object)
        self.amount = NumParam(info='the amount to apply', mandatory=True,)

        self.rand = NumParam(info='use uniform ramdom sampling', default=0)
        self.lb = NumParam(info='lower bound of random sampling', default=0)
        self.ub = NumParam(info='upper bound of random sampling', default=0)


class AlterModel(Model):
    """
    Implementation of the Alter model.
    """

    def __init__(self, system, config):
        Model.__init__(self, system, config)
        self.flags.tds = True
        self.group = 'TimedEvent'

        self.SW = Switcher(u=self.method, options=('+', '-', '*', '/', '='),
                           info='Switcher for alteration method',
                           )

        self.t.callback = self._alter_field

    def _alter_field(self, is_time):
        """
        Actuation of the alteration.
        """
        action = False

        for ii in range(self.n):
            if (not is_time[ii]) or (self.u.v[ii] == 0):
                continue

            model = self.system.__dict__[self.model.v[ii]]
            idx = self.dev.v[ii]
            src = self.src.v[ii]
            attr = self.attr.v[ii]
            amount = self.amount.v[ii]

            if self.rand.v[ii] == 1:
                amount = np.random.uniform(low=self.lb.v[ii], high=self.ub.v[ii])

            try:
                v0 = model.get(src=src, idx=idx, attr=attr)
            except KeyError as e:
                tqdm.write("\nError: <%s %s> cannot find idx=%s or src=%s in model <%s>. " % (
                    self.class_name, self.idx.v[ii],
                    idx, src, self.model.v[ii],
                ))
                tqdm.write("<%s %s> disabled due to %s.\n" %
                           (self.class_name, self.idx.v[ii], repr(e)))
                self.u.v[ii] = 0
                continue

            vnew = v0
            if self.SW.s0[ii] == 1:
                vnew = v0 + amount
            elif self.SW.s1[ii] == 1:
                vnew = v0 - amount
            elif self.SW.s2[ii] == 1:
                vnew = v0 * amount
            elif self.SW.s3[ii] == 1:
                vnew = v0 / amount
            elif self.SW.s4[ii] == 1:
                vnew = amount
            else:
                tqdm.write('Error: <%s %s>: undefined method "%s". <%s, %s> disabled.' % (
                    self.class_name, self.idx.v[ii], self.method.v[ii],
                    self.class_name, self.idx.v[ii]
                ))
                self.u.v[ii] = 0
                continue

            model.set(src=src, idx=idx, attr=attr, value=vnew)
            tqdm.write('<Alter %s>: set %s.%s.%s.%s=%.6g at t=%.6g. Previous value was %.6g.' % (
                self.idx.v[ii], self.model.v[ii], idx, src, attr, vnew, self.t.v[ii], v0
            ))
            action = True

        return action


class Alter(AlterData, AlterModel):
    """
    Model for altering device internal data at predefined time.

    Alter is useful to apply load changing, tap changing, step response, etc.
    can be applied to parameters and constant services but cannot be used to
    update variables.

    Alter is implemented by applying the given calculation to the ``v`` field of
    the linked parameter or constant. Alter will not affect other parameters or
    constants that depend on the altered variable.

    It is not uncommon for equations to depend on intermediate constants rather
    than the input parameters. Therefore, one will need to inspect model
    equations to determine the parameter/service to be altered.

    Examples
    --------
    To apply a PQ load change, according to :ref:`PQ`, one needs to set the load
    model to constant power and alter ``Ppf`` and ``Qpf``. Altering ``p0`` and
    ``q0`` will have no impact as they are not used in the equations for
    time-domain simulation.

    """

    def __init__(self, system, config):
        AlterData.__init__(self)
        AlterModel.__init__(self, system, config)



import logging
import inspect
from collections import OrderedDict

import numpy as np

from andes.core.service import BackRef
from andes.utils.func import list_flatten

logger = logging.getLogger(__name__)


class GroupBase:
    """
    Base class for groups.
    """

    def __init__(self):
        self.common_params = ['u', 'name']
        self.common_vars = []

        self.models = OrderedDict()        # model name: model instance
        self._idx2model = OrderedDict()    # element idx: model instance
        self.uid = {}                      # idx - group internal 0-indexed uid
        self.services_ref = OrderedDict()  # BackRef

    def __setattr__(self, key, value):
        if hasattr(value, 'owner'):
            if value.owner is None:
                value.owner = self
        if hasattr(value, 'name'):
            if value.name is None:
                value.name = key

        if isinstance(value, BackRef):
            self.services_ref[key] = value

        super().__setattr__(key, value)

    @property
    def class_name(self):
        return self.__class__.__name__

    @property
    def n(self):
        """
        Total number of devices.
        """
        return len(self._idx2model)

    def add_model(self, name: str, instance):
        """
        Add a Model instance to group.

        Parameters
        ----------
        name : str
            Model name
        instance : Model
            Model instance

        Returns
        -------
        None
        """
        if name not in self.models:
            self.models[name] = instance
        else:
            raise KeyError(f"{self.class_name}: Duplicate model registration of {name}")

    def add(self, idx, model):
        """
        Register an idx from model_name to the group

        Parameters
        ----------
        idx: Union[str, float, int]
            Register an element to a model

        model: Model
            instance of the model

        Returns
        -------

        """
        if idx in self._idx2model:
            raise KeyError(f'Group <{self.class_name}> already contains <{repr(idx)}> from '
                           f'<{self._idx2model[idx].class_name}>')
        self.uid[idx] = self.n
        self._idx2model[idx] = model

    def idx2model(self, idx, allow_none=False):
        """
        Find model name for the given idx.

        Parameters
        ----------
        idx : float, int, str, array-like
            idx or idx-es of devices.
        allow_none : bool
           If True, return `None` at the positions where idx is not found.

        Returns
        -------
        If `idx` is a list, return a list of model instances.
        If `idx` is a single element, return a model instance.
        """

        ret = []
        idx, single = self._1d_vectorize(idx)

        for i in idx:
            try:
                if i is None and allow_none:
                    ret.append(None)
                else:
                    ret.append(self._idx2model[i])
            except KeyError:
                raise KeyError(f'Group <{self.class_name}> does not contain device with idx={i}')

        if single:
            ret = ret[0]
        return ret

    def idx2uid(self, idx):
        """
        Convert idx to the 0-indexed unique index.

        Parameters
        ----------
        idx : array-like, numbers, or str
            idx of devices

        Returns
        -------
        list
            A list containing the unique indices of the devices
        """
        vec_idx, single = self._1d_vectorize(idx)

        out = [self.uid[i] if i is not None else None for i in vec_idx]
        if single:
            out = out[0]

        return out

    def get(self, src: str, idx, attr: str = 'v', allow_none=False, default=0.0):
        """
        Based on the indexer, get the `attr` field of the `src` parameter or variable.

        Parameters
        ----------
        src : str
            param or var name
        idx : array-like
            device idx
        attr
            The attribute of the param or var to retrieve
        allow_none : bool
            True to allow None values in the indexer
        default : float
            If `allow_none` is true, the default value to use for None indexer.

        Returns
        -------
        The requested param or variable attribute. If `idx` is a list, return a list of values.
        If `idx` is a single element, return a single value.
        """
        self._check_src(src)
        self._check_idx(idx)
        idx, single = self._1d_vectorize(idx)

        n = len(idx)
        if n == 0:
            return np.zeros(0)

        ret = [''] * n
        _type_set = False

        models = self.idx2model(idx, allow_none=allow_none)

        for i, idx in enumerate(idx):
            if models[i] is not None:
                uid = models[i].idx2uid(idx)
                instance = models[i].__dict__[src]
                val = instance.__dict__[attr][uid]
            else:
                val = default

            # deduce the type for ret
            if not _type_set:
                if isinstance(val, str):
                    ret = [''] * n
                else:
                    ret = np.zeros(n)
                _type_set = True

            ret[i] = val

        if single:
            ret = ret[0]

        return ret

    def set(self, src: str, idx, attr, value):
        """
        Set the value of an attribute of a group property.
        Performs ``self.<src>.<attr>[idx] = value``.

        The user needs to ensure that the property is shared by all models
        in this group.

        Parameters
        ----------
        src : str
            Name of property.
        idx : str, int, float, array-like
            Indices of devices.
        attr : str, optional, default='v'
            The internal attribute of the property to get.
            ``v`` for values, ``a`` for address, and ``e`` for equation value.
        value : array-like
            New values to be set

        Returns
        -------
        bool
            True when successful.
        """
        self._check_src(src)
        self._check_idx(idx)

        idx, _ = self._1d_vectorize(idx)
        models = self.idx2model(idx)

        if isinstance(value, (str, int, float, np.integer, np.floating)):
            value = [value] * len(idx)

        for mdl, ii, val in zip(models, idx, value):
            uid = mdl.idx2uid(ii)
            mdl.__dict__[src].__dict__[attr][uid] = val

        return True

    def find_idx(self, keys, values, allow_none=False, default=None):
        """
        Find indices of devices that satisfy the given `key=value` condition.

        This method iterates over all models in this group.
        """
        indices_found = []
        # `indices_found` contains found indices returned from all models of this group
        for model in self.models.values():
            indices_found.append(model.find_idx(keys, values, allow_none=True, default=default))

        out = []
        for idx, idx_found in enumerate(zip(*indices_found)):
            if not allow_none:
                if idx_found.count(None) == len(idx_found):
                    missing_values = [item[idx] for item in values]
                    raise IndexError(f'{list(keys)} = {missing_values} not found in {self.class_name}')

            real_idx = default
            for item in idx_found:
                if item is not None:
                    real_idx = item
                    break
            out.append(real_idx)
        return out

    def _check_src(self, src: str):
        """
        Helper function for checking if ``src`` is a shared field.

        The requirement is not strictly enforced and is only for debugging purposed.
        """
        if src not in self.common_vars + self.common_params:
            logger.debug(f'Group <{self.class_name}> does not share property <{src}>.')

    def _check_idx(self, idx):
        """
        Helper function for checking if ``idx`` is None.

        Raises IndexError if idx is None.
        """

        if idx is None:
            raise IndexError(f'{self.class_name}: idx cannot be None')

    def _1d_vectorize(self, idx):
        """
        Helper function to convert a single element, list, or nested lists
        into a list.

        If the input is a nested list, flatten it into a 1-dimensional
        list.

        Returns
        -------
        idx : list
            List of indices.
        single : bool
            True if the input is a single element.
        """
        single = False
        list_alike = (list, tuple, np.ndarray)

        if not isinstance(idx, list_alike):
            idx = [idx]
            single = True
        elif len(idx) > 0 and isinstance(idx[0], list_alike):
            idx = list_flatten(idx)

        return idx, single

    def get_field(self, src: str, idx, field: str):
        """
        Helper function for retrieving an attribute of a member variable shared
        by models in this group.

        Returns
        -------
        list
            A list with the length equal to ``len(idx)``.
        """

        self._check_src(src)
        self._check_idx(idx)

        idx, _ = self._1d_vectorize(idx)
        models = self.idx2model(idx, allow_none=True)

        ret = [None] * len(models)
        for ii, model in enumerate(models):
            if model is not None:
                ret[ii] = getattr(model.__dict__[src], field)

        return ret

    def set_backref(self, name, from_idx, to_idx):
        """
        Set idxes to ``BackRef``, and set them to models.
        """

        uid = self.idx2uid(to_idx)
        self.services_ref[name].v[uid].append(from_idx)

        model = self.idx2model(to_idx)
        model.set_backref(name, from_idx, to_idx)

    def get_next_idx(self, idx=None, model_name=None):
        """
        Get a no-conflict idx for a new device.
        Use the provided ``idx`` if no conflict.
        Generate a new one otherwise.

        Parameters
        ----------
        idx : str or None
            Proposed idx. If None, assign a new one.
        model_name : str or None
            Model name. If not, prepend the group name.

        Returns
        -------
        str
            New device name.

        """
        if model_name is None:
            model_name = self.class_name

        need_new = False

        if idx is not None:
            if idx not in self._idx2model:
                # name is good
                pass
            else:
                logger.warning("Group <%s>: idx=%s is used by %s. Data may be inconsistent.",
                               self.class_name, idx, self.idx2model(idx).class_name)
                need_new = True
        else:
            need_new = True

        if need_new is True:
            count = self.n
            while True:
                # IMPORTANT: automatically assigned index is 1-indexed. Namely, `GENCLS_1` is the first generator.
                # This is because when we say, for example, `GENCLS_10`, people usually assume it starts at 1.
                idx = model_name + '_' + str(count + 1)
                if idx not in self._idx2model:
                    break
                else:
                    count += 1

        return idx

    def doc(self, export='plain'):
        """
        Return the documentation of the group in a string.
        """
        out = ''
        if export == 'rest':
            out += f'.. _{self.class_name}:\n\n'
            group_header = '=' * 80 + '\n'
        else:
            group_header = ''

        if export == 'rest':
            out += group_header + f'{self.class_name}\n' + group_header
        else:
            out += group_header + f'Group <{self.class_name}>\n' + group_header

        if self.__doc__ is not None:
            out += inspect.cleandoc(self.__doc__) + '\n\n'

        if len(self.common_params):
            out += 'Common Parameters: ' + ', '.join(self.common_params)
            out += '\n\n'
        if len(self.common_vars):
            out += 'Common Variables: ' + ', '.join(self.common_vars)
            out += '\n\n'
        if len(self.models):
            out += 'Available models:\n'
            model_name_list = list(self.models.keys())

            if export == 'rest':
                def add_reference(name_list):
                    return [f'{item}_' for item in name_list]

                model_name_list = add_reference(model_name_list)

            out += ',\n'.join(model_name_list) + '\n'

        return out

    def doc_all(self, export='plain'):
        """
        Return documentation of the group and its models.

        Parameters
        ----------
        export : 'plain' or 'rest'
            Export format, plain-text or RestructuredText

        Returns
        -------
        str

        """
        out = self.doc(export=export)
        out += '\n'
        for instance in self.models.values():
            out += instance.doc(export=export)
            out += '\n'
        return out


class Undefined(GroupBase):
    """
    The undefined group. Holds models with no ``group``.
    """
    pass


class ACTopology(GroupBase):
    def __init__(self):
        super().__init__()
        self.common_vars.extend(('a', 'v'))


class DCTopology(GroupBase):
    def __init__(self):
        super().__init__()
        self.common_vars.extend(('v',))


class Collection(GroupBase):
    """Collection of topology models"""
    pass


class Calculation(GroupBase):
    """Group of classes that calculates based on other models."""
    pass


class StaticGen(GroupBase):
    """
    Static generator group.

    Static generators will be replaced by dynamic generators, either synchronous
    generators or inverter-based resources upon the initialization for dynamics.
    It is implemented by setting the connectivity status ``u`` of the replaced
    StaticGen to 0.

    See the notes in :ref:`SynGen` for replacing one StaticGen with multiple
    dynamic ones.
    """

    def __init__(self):
        super().__init__()
        self.common_params.extend(('Sn', 'Vn', 'p0', 'q0', 'ra', 'xs', 'subidx'))
        self.common_vars.extend(('q', 'a', 'v'))

        self.SynGen = BackRef()


class ACLine(GroupBase):
    def __init__(self):
        super(ACLine, self).__init__()
        self.common_params.extend(('bus1', 'bus2', 'r', 'x'))
        self.common_vars.extend(('v1', 'v2', 'a1', 'a2'))


class ACShort(GroupBase):
    def __init__(self):
        super(ACShort, self).__init__()
        self.common_params.extend(('bus1', 'bus2'))
        self.common_vars.extend(('v1', 'v2', 'a1', 'a2'))


class StaticLoad(GroupBase):
    """
    Static load group.
    """
    pass


class StaticShunt(GroupBase):
    """
    Static shunt compensator group.
    """
    pass


class DynLoad(GroupBase):
    """
    Dynamic load group.
    """
    pass


class SynGen(GroupBase):
    """
    Synchronous generator group.

    SynGen replaces StaticGen upon the initialization of dynamic studies. SynGen
    and inverter-based resources contain parameters ``gammap`` and ``gammaq``
    for splitting the initial power of a StaticGen into multiple dynamic ones.

    ``gammap``, for example, is the active power ratio of the dynamic generator
    to the static one. If a StaticGen is supposed to be replaced by one SynGen,
    the ``gammap`` and ``gammaq`` should both be ``1``.

    It is critical to ensure that ``gammap`` and ``gammaq``, respectively, of
    all dynamic power sources sum up to 1.0. Otherwise, the initial power
    injections imposed by dynamic sources will differ from the static ones. The
    initialization will then fail with mismatches power injection equations
    corresponding to bus ``a`` and ``v``.

    """

    def __init__(self):
        super().__init__()
        self.common_params.extend(('Sn', 'Vn', 'fn', 'bus', 'M', 'D', 'subidx'))
        self.common_vars.extend(('omega', 'delta', ))
        self.idx_island = []
        self.uid_island = []
        self.delta_addr = []

        self.TurbineGov = BackRef()
        self.Exciter = BackRef()

    def store_idx_island(self, bus_idx):
        """
        Get ``idx`` of generators in the given islanded. Also store the
        addresses of the ``delta`` variable in the largest island.

        This function can only be called after initializing dynamic devices.

        Parameters
        ----------
        bus_idx : list
            A list of bus idx in the largest island
        """

        idx_gen = list(self.uid.keys())
        bus_gen = self.get(src='bus', idx=idx_gen, attr='v')

        def intersect(lst1, lst2):
            return list(set(lst1) & set(lst2))

        bus_gen_island = intersect(bus_idx, bus_gen)
        self.idx_island = self.find_idx(keys='bus',
                                        values=bus_gen_island)
        self.uid_island = self.idx2uid(self.idx_island)

        self.delta_addr = self.get('delta', self.idx_island, 'a').astype(int)


class RenGen(GroupBase):
    """
    Renewable generator (converter) group.

    See :ref:`SynGen` for the notes on replacing StaticGen and setting the power
    ratio parameters.
    """

    def __init__(self):
        super().__init__()
        self.common_params.extend(('bus', 'gen', 'Sn'))
        self.common_vars.extend(('Pe', 'Qe'))


class RenExciter(GroupBase):
    """
    Renewable electrical control (exciter) group.
    """

    def __init__(self):
        super().__init__()
        self.common_params.extend(('reg',))
        self.common_vars.extend(('Pref', 'Qref', 'wg', 'Pord'))


class RenPlant(GroupBase):
    """
    Renewable plant control group.
    """

    def __init__(self):
        super().__init__()


class RenGovernor(GroupBase):
    """
    Renewable turbine governor group.
    """

    def __init__(self):
        super().__init__()
        self.common_params.extend(('ree', 'w0', 'Sn', 'Pe0'))
        self.common_vars.extend(('Pm', 'wr0', 'wt', 'wg', 's3_y'))


class RenAerodynamics(GroupBase):
    """
    Renewable aerodynamics group.
    """

    def __init__(self):
        super().__init__()
        self.common_params.extend(('rego',))
        self.common_vars.extend(('theta',))


class RenPitch(GroupBase):
    """
    Renewable generator pitch controller group.
    """

    def __init__(self):
        super().__init__()
        self.common_params.extend(('rea',))


class RenTorque(GroupBase):
    """
    Renewable torque (Pref) controller.
    """

    def __init__(self):
        super().__init__()


class DG(GroupBase):
    """
    Distributed generation (small-scale).

    See :ref:`SynGen` for the notes on replacing StaticGen and setting the power
    ratio parameters.
    """

    def __init__(self):
        super().__init__()
        self.common_params.extend(('bus', 'fn'))


class DGProtection(GroupBase):
    """
    Protection model for DG.
    """

    def __init__(self):
        super().__init__()


class TurbineGov(GroupBase):
    """
    Turbine governor group for synchronous generator.
    """

    def __init__(self):
        super().__init__()
        self.common_vars.extend(('pout',))


class Exciter(GroupBase):
    """
    Exciter group for synchronous generators.
    """

    def __init__(self):
        super().__init__()
        self.common_params.extend(('syn',))
        self.common_vars.extend(('vout', 'vi',))

        self.VoltComp = BackRef()
        self.PSS = BackRef()


class VoltComp(GroupBase):
    """
    Voltage compensator group for synchronous generators.
    """

    def __init__(self):
        super().__init__()
        self.common_params.extend(('rc', 'xc',))
        self.common_vars.extend(('vcomp',))


class PSS(GroupBase):
    """Power system stabilizer group."""

    def __init__(self):
        super().__init__()
        self.common_vars.extend(('vsout',))


class Experimental(GroupBase):
    """Experimental group"""
    pass


class DCLink(GroupBase):
    """Basic DC links"""
    pass


class StaticACDC(GroupBase):
    """AC DC device for power flow"""
    pass


class TimedEvent(GroupBase):
    """Timed event group"""
    pass


class FreqMeasurement(GroupBase):
    """Frequency measurements."""

    def __init__(self):
        super().__init__()
        self.common_vars.extend(('f',))


class PhasorMeasurement(GroupBase):
    """Phasor measurements"""

    def __init__(self):
        super().__init__()
        self.common_vars.extend(('am', 'vm'))


class PLL(GroupBase):
    """Phase-locked loop models."""

    def __init__(self):
        super().__init__()
        self.common_vars.extend(('am',))


class Motor(GroupBase):
    """Induction Motor group
    """

    def __init__(self):
        super().__init__()


class Information(GroupBase):
    """
    Group for information container models.
    """

    def __init__(self):
        GroupBase.__init__(self)
        self.common_params = []


class OutputSelect(GroupBase):
    """
    Group for selecting outputs.
    """

    def __init__(self):
        super().__init__()
        self.common_params = []


class Interface(GroupBase):
    """
    Group for interface models.
    """

    def __init__(self):
        super().__init__()
        self.common_params = []


class DataSeries(GroupBase):
    """
    Group for TimeSeries models.
    """

    def __init__(self):
        super().__init__()
        self.common_params = []



"""
Renewable energy generator (converter) model P with PLL support.

This is a customized model in ANDES, modified from the standard REGCA1 model.
"""

from andes.core import Algeb, ExtState, IdxParam
from andes.core.service import DeviceFinder, FlagValue
from andes.models.renewable.regca1 import REGCA1Data, REGCA1Model


class REGCP1Data(REGCA1Data):
    """
    REGCP1 model data.
    """

    def __init__(self):
        REGCA1Data.__init__(self)

        self.pll = IdxParam(info='Phase-lock loop device idx',
                            model='PLL',
                            default=None,
                            )


class REGCP1Model(REGCA1Model):
    """
    REGCP1 implementation.
    """

    def __init__(self, system, config):
        REGCA1Model.__init__(self, system, config)
        self.pllidx = DeviceFinder(self.pll,
                                   link=self.bus,
                                   idx_name='bus',
                                   default_model='PLL1',
                                   auto_find=False,
                                   auto_add=False,
                                   )

        self.am = ExtState(model='PLL',
                           src='am',
                           indexer=self.pllidx,
                           allow_none=True,
                           tex_name=r'\theta_m',
                           info='Measured angle',
                           )

        self.zam = FlagValue(self.pllidx,
                             value=None,
                             flag=0.0,
                             tex_name='z_{\theta m}',
                             info='Flag invalid angle measurements as zeros',
                             )

        delattr(self, 'vd')
        self.algebs_ext.pop("vd")

        self.vd = Algeb(v_str='v',
                        info='d-axis voltage', tex_name='V_d',
                        e_str='v*cos(zam * (a - am)) - vd')

        self.vq = Algeb(v_str='0',
                        info='q-axis voltage', tex_name='V_q',
                        e_str='- v*sin(zam * (a - am)) - vq')

        self.Pe.e_str = '(vd * Ipout + vq * Iqout_y) - Pe'

        self.Qe.e_str = '(vd * Iqout_y - vq * Ipout) - Qe'

    def v_numeric(self, **kwargs):
        """
        Disable the corresponding `StaticGen`s.
        """
        self.system.groups['StaticGen'].set(src='u', idx=self.gen.v, attr='v', value=0)


class REGCP1(REGCP1Data, REGCP1Model):
    """
    Renewable energy generator model (REGC_A) with PLL.

    A PLL device can be specified for estimating the phase angle at the coupling
    bus through the ``pll`` parameter:

    - If ``pll`` is not given, the accurate bus angle will be used.

    - If ``pll`` is not a valid PLL device, the program will error out.

    - The program does not check if the provided ``pll`` actually measures the
      bus on which the converter is at.

    One needs to carefully tune the PLL parameters to match the desired
    performance.

    All other remarks for ``REGCA1`` apply.
    """

    def __init__(self, system, config):
        REGCP1Data.__init__(self)
        REGCP1Model.__init__(self, system, config)



"""
ESDC1A model.
"""

from andes.core.service import ConstService

from andes.models.exciter.esdc2a import ESDC2AData, ESDC2AModel


class ESDC1AModel(ESDC2AModel):
    """
    Implementation of ESDC1A.
    """

    def __init__(self, system, config):
        ESDC2AModel.__init__(self, system, config)

        self.services.pop("VRU")
        self.services.pop("VRL")
        self.services_var.pop("VRU")
        self.services_var.pop("VRL")
        self.services_var_seq.pop("VRU")
        self.services_var_seq.pop("VRL")

        delattr(self, 'VRU')
        delattr(self, 'VRL')

        self.VRU = ConstService(v_str='VRMAXc',
                                tex_name='V_{RMAX}',
                                )
        self.VRL = ConstService(v_str='VRMIN',
                                tex_name='V_{RMIN}',
                                )


class ESDC1A(ESDC2AData, ESDC1AModel):
    """
    ESDC1A model.

    This model derives from ESDC2A and changes the regular limits to "VRMAX" and
    "VRMIN".

    """

    def __init__(self, system, config):
        ESDC2AData.__init__(self)
        ESDC1AModel.__init__(self, system, config)



"""
Test the ANDES-gridcal interface.
"""

import unittest

import andes
from andes.interop.gridcal import to_gridcal, _verify_pf

try:
    import GridCal.Engine as gc
    getattr(gc, '__name__')
    HAVE_GRIDCAL = True
except (ImportError, AttributeError):
    HAVE_GRIDCAL = False


@unittest.skipUnless(HAVE_GRIDCAL, "gridcal not available")
class TestGridcal(unittest.TestCase):
    """
    Tests for the ANDES-gridcal interface.
    """

    cases = [
        'ieee14/ieee14_ieeet1.xlsx',
        'ieee14/ieee14_pvd1.xlsx',
        'ieee39/ieee39.xlsx',
        'npcc/npcc.xlsx',
    ]

    def setUp(self) -> None:
        """
        Test setup. This is executed before each test case.
        """

    def test_to_gridcal(self):
        """
        Test `andes.interop.gridcal.to_gridcal` with cases
        """
        for case_file in self.cases:
            case = andes.get_case(case_file)
            match = _test_to_gridcal_single(case)

            self.assertEqual(match, True)


def _test_to_gridcal_single(case, **kwargs):
    """
    Test `andes.interop.gridcal.to_gridcal` with a single case
    """

    sa = andes.load(case, setup=True, no_output=True, default_config=True)
    sgc = to_gridcal(sa, verify=False, **kwargs)
    return _verify_pf(sa, sgc)



"""
Grid-forming renewable energy model with dispatchable virtual oscillator control
(dVOC).
"""

from andes.core.var import Algeb, State
from andes.core.service import ConstService
from andes.core.block import GainLimiter
from andes.models.renewable.regf1 import REGF1Data, REGF1Model, REGFInnerPIModel, REGFOuterPIModel


class REGF3Primary:
    """
    Primary frequency and voltage controllers based on dVOC.
    """

    def __init__(self) -> None:

        self.Kdvoc = ConstService(
            v_str='wdrp * (4 * 100000000) / (100000000 - (2*(100-100*Qdrp)**2 ) - 10000)**2',
            info='Kdvoc parameter',
        )

        self.wref = Algeb(v_str='1', e_str='1 - wref',
                          tex_name=r'\omega_{ref}', info='speed ref', unit='pu')

        self.Vref = Algeb(v_str='vref', e_str='vref - Vref')

        self.dw = GainLimiter(u='w0 * (PIplim_y - Psen_y) * wdrp / vref2 / vref2', K=1, R=1,
                              lower=self.dwmin,
                              upper=self.dwmax,
                              )

        derv = 'wdrp / vref2 * (PIqlim_y - Qsen_y) + (vref2 * Kdvoc * (Vref + vref2) * (Vref - vref2)) - derv'
        self.derv = Algeb(tex_name=r'dV',
                          info='input to voltage integrator',
                          e_str=derv,
                          v_str='0')

        self.vref2 = State(e_str='w0 * derv', v_str='vd')


class REGF3(REGF1Data, REGF1Model, REGF3Primary,
            REGFOuterPIModel, REGFInnerPIModel):
    """
    Grid-forming inverter with dVOC.

    Implementation of EPRI Memorandum

    D. Ramasubramanian, "PROPOSAL FOR SUITE OF GENERIC GRID FORMING (GFM) POSITIVE SEQUENCE MODELS"
    """

    def __init__(self, system, config):
        REGF1Data.__init__(self)

        REGF1Model.__init__(self, system, config)
        REGF3Primary.__init__(self)
        REGFOuterPIModel.__init__(self)
        REGFInnerPIModel.__init__(self)



"""
Module for ESAC5A exciter model.
"""

from andes.core.block import Lag, LagAntiWindup, LeadLag, Washout
from andes.core.param import NumParam
from andes.core.service import PostInitService, ConstService
from andes.core.var import Algeb
from andes.models.exciter.excbase import ExcACSat, ExcBase, ExcBaseData, ExcVsum


class ESAC5AData(ExcBaseData):
    """
    ESAC5A data.
    """

    def __init__(self):
        ExcBaseData.__init__(self)
        self.TR = NumParam(info='Sensing Time Constant',
                           tex_name='T_R',
                           default=0.01,
                           unit='p.u',
                           )

        self.TA = NumParam(info='Voltage Regulator Time Constant',
                           tex_name='T_A',
                           default=0.04,
                           unit='p.u',
                           )

        self.KA = NumParam(info='Voltage Regulator Gain',
                           tex_name='K_A',
                           default=80,
                           )

        self.VRMIN = NumParam(info='V_R lower limit',
                              default=-7.3,
                              unit='p.u',
                              tex_name='V_Rmin'
                              )

        self.VRMAX = NumParam(info='V_R upper limit',
                              default=7.3,
                              unit='p.u',
                              tex_name='V_Rmax',
                              )

        self.TE = NumParam(info='Integrator Time Constant',
                           tex_name='T_E',
                           default=0.8,
                           unit='p.u',
                           non_negative=True
                           )

        self.KF = NumParam(info='Feedback Gain',
                           default=0.03,
                           tex_name='K_F'
                           )

        self.TF1 = NumParam(info='Lag Time Constant',
                            default=1.0,
                            unit='p.u',
                            tex_name='T_F_1',
                            )

        self.TF2 = NumParam(info='Lead-Lag Time Constant (pole)',
                            default=0.8,
                            unit='p.u',
                            tex_name='T_F_2',
                            )

        self.TF3 = NumParam(info='Lead-Lag Time Constant (zero)',
                            default=1,
                            unit='p.u',
                            tex_name='T_F_3',
                            )

        self.KE = NumParam(info='Exciter Feedback Gain',
                           tex_name='K_E',
                           default=1,
                           )

        self.E1 = NumParam(info='First saturation point',
                           default=0,
                           unit='p.u.',
                           tex_name='E_1',
                           )
        self.SE1 = NumParam(info='Value at first saturation point',
                            default=0,
                            unit='p.u.',
                            tex_name='S_E1'
                            )
        self.E2 = NumParam(info='Second saturation point',
                           default=1,
                           unit='p.u.',
                           tex_name='E_2'
                           )
        self.SE2 = NumParam(info='Value at second saturation point',
                            default=1,
                            unit='p.u.',
                            tex_name='S_E2'
                            )


class ESAC5AModel(ExcBase, ExcVsum, ExcACSat):
    """
    ESAC5A model implementation.
    """

    def __init__(self, system, config):
        ExcBase.__init__(self, system, config)
        ExcVsum.__init__(self)

        self.LP = Lag(u=self.v, T=self.TR, K=1, info='Voltage transducer',)

        self.vi = Algeb(info='Total voltage input',
                        unit='pu',
                        e_str='ue * (-LP_y + vref + Vs - WF_y ) -vi ',
                        v_str='ue*(-v +vref)',
                        )

        self.VRMAXu = ConstService('VRMAX * ue + (1-ue) * 999')
        self.VRMINu = ConstService('VRMIN * ue + (1-ue) * -999')

        self.VR = LagAntiWindup(u=self.vi, T=self.TA, K=self.KA, upper=self.VRMAXu, lower=self.VRMINu,)

        self.LL = LeadLag(u=self.VR_y, T1=self.TF3, T2=self.TF2,)

        self.WF = Washout(u=self.LL_y, T=self.TF1, K=self.KF)

        self.INTin = 'ue * (VR_y - VFE)'

        ExcACSat.__init__(self)

        self.vref.v_str = 'v + VFE / KA'

        self.vref0 = PostInitService(info='Initial reference voltage input',
                                     tex_name='V_{ref0}',
                                     v_str='vref',
                                     )

        self.VFE.v_str = "INT_y * KE + Se "
        self.VFE.e_str = "ue * (INT_y * KE + Se - VFE) "

        # disable iterative initialization of the integrator output
        self.INT.y.v_str = 'vf0'
        self.INT.y.v_iter = None

        self.vout.e_str = 'ue * INT_y - vout'


class ESAC5A (ESAC5AData, ESAC5AModel):
    """
    Exciter ESAC5A.
    """

    def __init__(self, system, config):
        ESAC5AData.__init__(self)
        ESAC5AModel.__init__(self, system, config)



"""
Simple pandapower (2.7.0) interface
"""

import logging
import numpy as np
from functools import wraps

from andes.shared import pd, rad2deg, deg2rad
from andes.shared import pandapower as pp

logger = logging.getLogger(__name__)


def require_pandapower(f):
    """
    Decorator for functions that require pandapower.
    """

    @wraps(f)
    def wrapper(*args, **kwds):
        try:
            getattr(pp, '__version__')
        except AttributeError:
            raise ModuleNotFoundError("pandapower needs to be manually installed.")

        return f(*args, **kwds)

    return wrapper


def build_group_table(ssa, group, columns, mdl_name=[]):
    """
    Build the table for devices in a group in an ANDES System.

    Parameters
    ----------
    ssa : andes.system.System
        The ANDES system to build the table
    group : string
        The ANDES group
    columns : list of string
        The common columns of a group that to be included in the table.
    mdl_name : list of string
        The list of models that to be included in the table. Default as all models.

    Returns
    -------
    DataFrame

        The output Dataframe contains the columns from the device
    """
    group_df = pd.DataFrame(columns=columns)
    group = getattr(ssa, group)
    if not mdl_name:
        mdl_dict = getattr(group, 'models')
        for key in mdl_dict:
            mdl = getattr(ssa, key)
            group_df = pd.concat([group_df, mdl.as_df()[columns]], axis=0)
    else:
        for key in mdl_name:
            mdl = getattr(ssa, key)
            group_df = pd.concat([group_df, mdl.as_df()[columns]], axis=0)
    return group_df.reset_index(drop=True)


def make_link_table(ssa):
    """
    Build the link table for generators and generator controllers in an ANDES
    System, including ``SynGen`` and ``DG`` for now.

    Parameters
    ----------
    ssa : andes.system.System
        The ANDES system to link

    Returns
    -------
    DataFrame

        Each column in the output Dataframe contains the ``idx`` of linked
        ``StaticGen``, ``Bus``, ``DG``, ``RenGen``, ``RenExciter``, ``SynGen``,
        ``Exciter``, and ``TurbineGov``, ``gammap``, ``gammaq``.
    """
    # build StaticGen df
    ssa_stg = build_group_table(ssa, 'StaticGen', ['u', 'name', 'idx', 'bus'])
    # build TurbineGov df
    ssa_gov = build_group_table(ssa, 'TurbineGov', ['idx', 'syn'])
    # build Exciter df
    ssa_exc = build_group_table(ssa, 'Exciter', ['idx', 'syn'])
    # build SynGen df
    ssa_syg = build_group_table(ssa, 'SynGen', ['idx', 'bus', 'gen', 'gammap', 'gammaq'], ['GENCLS', 'GENROU'])
    # build DG df
    ssa_dg = build_group_table(ssa, 'DG', ['idx', 'bus', 'gen', 'gammap', 'gammaq'])
    # build RenGen df
    ssa_rg = build_group_table(ssa, 'RenGen', ['idx', 'bus', 'gen', 'gammap', 'gammaq'])
    # build RenExciter df
    ssa_rexc = build_group_table(ssa, 'RenExciter', ['idx', 'reg'])

    # output
    ssa_bus = ssa.Bus.as_df()[['name', 'idx']]
    ssa_key = pd.merge(left=ssa_stg.rename(columns={'name': 'stg_name', 'idx': 'stg_idx',
                                                    'bus': 'bus_idx', 'u': 'stg_u'}),
                       right=ssa_bus.rename(columns={'name': 'bus_name', 'idx': 'bus_idx'}),
                       how='left', on='bus_idx')
    ssa_syg = pd.merge(left=ssa_key, how='right', on='stg_idx',
                       right=ssa_syg.rename(columns={'idx': 'syg_idx', 'gen': 'stg_idx'}))
    ssa_dg = pd.merge(left=ssa_key, how='right', on='stg_idx',
                      right=ssa_dg.rename(columns={'idx': 'dg_idx', 'gen': 'stg_idx'}))
    ssa_rg = pd.merge(left=ssa_key, how='right', on='stg_idx',
                      right=ssa_rg.rename(columns={'idx': 'rg_idx', 'gen': 'stg_idx'}))
    ssa_key0 = pd.merge(left=ssa_key, how='left', on='stg_idx',
                        right=ssa_syg[['stg_idx', 'syg_idx']])
    ssa_key0 = pd.merge(left=ssa_key0, how='left', on='stg_idx',
                        right=ssa_dg[['stg_idx', 'dg_idx']])
    ssa_key0 = pd.merge(left=ssa_key0, how='left', on='stg_idx',
                        right=ssa_rg[['stg_idx', 'rg_idx']])
    ssa_key0.fillna(False, inplace=True)
    ssa_key0['dyr'] = ssa_key0['syg_idx'].astype(
        bool) + ssa_key0['dg_idx'].astype(bool) + ssa_key0['rg_idx'].astype(bool)
    ssa_key0['dyr'] = 1 - ssa_key0['dyr'].astype(int)
    ssa_key0['dyr'] = ssa_key0['dyr'].astype(bool)
    ssa_dyr0 = ssa_key0[ssa_key0.dyr].drop(['dyr'], axis=1)
    ssa_dyr0['gammap'] = 1
    ssa_dyr0['gammaq'] = 1
    ssa_key = pd.concat([ssa_syg, ssa_dg, ssa_rg, ssa_dyr0], axis=0)
    ssa_key = pd.merge(left=ssa_key,
                       right=ssa_exc.rename(columns={'idx': 'exc_idx', 'syn': 'syg_idx'}),
                       how='left', on='syg_idx')
    ssa_key = pd.merge(left=ssa_key,
                       right=ssa_gov.rename(columns={'idx': 'gov_idx', 'syn': 'syg_idx'}),
                       how='left', on='syg_idx')
    ssa_key = pd.merge(left=ssa_key, how='left', on='rg_idx',
                       right=ssa_rexc.rename(columns={'idx': 'rexc_idx', 'reg': 'rg_idx'}))

    cols = ['stg_name', 'stg_u', 'stg_idx', 'bus_idx', 'dg_idx', 'rg_idx', 'rexc_idx',
            'syg_idx', 'exc_idx', 'gov_idx', 'bus_name', 'gammap', 'gammaq']
    return ssa_key[cols].reset_index(drop=True)


@require_pandapower
def runopp_map(ssp, link_table, **kwargs):
    """
    Run OPF in pandapower using ``pp.runopp()`` and map results back to ANDES
    based on the link table.

    Parameters
    ----------
    ssp : pandapower network
        The pandapower network

    link_table : DataFrame
        The link table of ANDES system

    Returns
    -------
    DataFrame

        The DataFrame contains the OPF results with columns ``p_mw``,
        ``q_mvar``, ``vm_pu`` in p.u., and the corresponding ``idx`` of
        ``StaticGen``, ``Exciter``, ``TurbineGov`` in ANDES.

    Notes
    -----
      - The pandapower net and the ANDES system must have same base MVA.
      - Multiple ``DG`` connected to the same ``StaticGen`` will be converted to one generator.
        The power is dispatched to each ``DG`` by the power ratio ``gammap``
    """

    try:
        pp.runopp(ssp, **kwargs)
        logger.warning("ACOPF is solved.")
    except Exception:
        pp.rundcopp(ssp, **kwargs)
        logger.warning("ACOPF failed. DCOPF is solved.")

    # take dispatch results from pp
    ssp_gen = ssp.gen.rename(columns={'name': 'stg_idx'})
    ssp_res = pd.concat([ssp_gen[['stg_idx']],
                        ssp.res_gen[['p_mw', 'q_mvar', 'vm_pu']]], axis=1)

    ssp_res = pd.merge(left=ssp_res,
                       right=ssp_gen[['stg_idx', 'controllable']],
                       how='left', on='stg_idx')
    ssp_res = pd.merge(left=ssp_res, right=link_table,
                       how='left', on='stg_idx')
    ssp_res['p'] = ssp_res['p_mw'] * ssp_res['gammap'] / ssp.sn_mva
    ssp_res['q'] = ssp_res['q_mvar'] * ssp_res['gammaq'] / ssp.sn_mva
    col = ['stg_idx', 'p', 'q', 'vm_pu', 'bus_idx', 'controllable',
           'dg_idx', 'rg_idx', 'syg_idx', 'gov_idx', 'exc_idx']
    return ssp_res[col]


@require_pandapower
def add_gencost(ssp, gen_cost):
    """
    Add cost function to converted pandapower net `ssp`.

    The cost data follows the same format of pypower and matpower.

    Now only poly_cost is supported.

    Parameters
    ----------
    ssp :
        The pandapower net

    gen_cost : array
        generator cost data
    """

    # check dimension
    if gen_cost.shape[0] != ssp.gen.shape[0]:
        print('Input cost function size does not match gen size.')

    for num, uid in enumerate(ssp.gen.index):
        if gen_cost[num, 0] == 2:  # poly_cost
            pp.create_poly_cost(net=ssp,
                                element=uid,
                                et='gen',
                                cp2_eur_per_mw2=gen_cost[num, 4],
                                cp1_eur_per_mw=gen_cost[num, 5],
                                cp0_eur=gen_cost[num, 6])
        else:
            # TODO: piecewise linear
            continue

    return True


def _to_pp_bus(ssp, ssa_bus):
    """Create bus in pandapower net"""
    bus_cols = ['Vn', 'name', 'u', 'vmax', 'vmin', 'zone']
    bus_df = ssa_bus[bus_cols]
    bus_df.columns = ['vn_kv', 'name', 'in_service', 'max_vm_pu', 'min_vm_pu', 'zone']
    bus_df['type'] = 'b'
    bus_df['in_service'] = bus_df['in_service'].astype('bool')
    bus_df.reset_index(inplace=True, drop=True)
    setattr(ssp, 'bus', bus_df)
    return ssp


def _to_pp_line(ssa, ssp, ssa_bus):
    """Create line in pandapower net"""
    # TODO: 1) from- and to- sides `Y`; 2)`g`
    omega = 2 * np.pi * ssp.f_hz

    ssa_bus_slice = ssa.Bus.as_df()[['idx', 'Vn']].rename(columns={"idx": "bus1", "Vn": "Vb"})
    ssa_line = ssa.Line.as_df().merge(ssa_bus_slice, on='bus1', how='left')

    ssa_line['Zb'] = ssa_line["Vb"]**2 / ssa_line["Sn"]
    ssa_line['Yb'] = ssa_line["Sn"] / ssa_line["Vb"]**2
    ssa_line['R'] = ssa_line["r"] * ssa_line['Zb']  # ohm
    ssa_line['X'] = ssa_line["x"] * ssa_line['Zb']  # ohm
    ssa_line['C'] = ssa_line["b"] / ssa_line['Zb'] / omega * 1e9  # nF
    ssa_line['G'] = ssa_line["g"] * ssa_line['Yb'] * 1e6  # mS
    # default rate_a is 2000 MVA
    ssa_line['rate_a'] = ssa_line['rate_a'].replace(0, 2000)
    ssa_line['max_i_ka'] = ssa_line["rate_a"] / ssa_line['Vb'] / np.sqrt(3)  # kA

    ssa_bus1 = ssa_bus[['idx']]
    ssa_bus1['from_bus'] = ssa_bus.index
    ssa_bus1.rename(columns={'uid': 'from_bus', 'idx': 'bus1'}, inplace=True)
    ssa_bus1 = ssa_bus1.reset_index(drop=True)
    ssa_bus2 = ssa_bus1.copy()
    ssa_bus2.rename(columns={'from_bus': 'to_bus', 'bus1': 'bus2'}, inplace=True)
    ssa_line = ssa_line.merge(ssa_bus1, on='bus1', how='left')
    ssa_line = ssa_line.merge(ssa_bus2, on='bus2', how='left')

    ssa_line['name'] = _rename(ssa_line['name'])

    line_cols = ['name', 'from_bus', 'to_bus', 'u', 'R', 'X', 'C', 'G', 'max_i_ka']
    line_df = ssa_line[line_cols][ssa_line['trans'] == 0].reset_index(drop=True)
    line_df['length_km'] = 1
    line_df['type'] = 'ol'
    line_df['parallel'] = 1
    line_df['max_loading_percent'] = 100
    line_df['df'] = 1
    line_df['std_type'] = None
    line_df.rename(columns={'R': 'r_ohm_per_km', 'X': 'x_ohm_per_km',
                            'C': 'c_nf_per_km', 'G': 'g_us_per_km',
                            'u': 'in_service'}, inplace=True)
    line_df['in_service'] = line_df['in_service'].astype('bool')
    setattr(ssp, 'line', line_df)

    tf_df = ssa_line[ssa_line['trans'] == 1].reset_index(drop=True)
    if tf_df.shape[0] > 1:
        tf_df.rename(columns={'R': 'r_ohm_per_km', 'X': 'x_ohm_per_km',
                              'C': 'c_nf_per_km', 'G': 'g_us_per_km',
                              'u': 'in_service'}, inplace=True)
        tf_df['in_service'] = tf_df['in_service'].astype('bool')

        tf_df['hv_bus'] = tf_df[['from_bus', 'to_bus', 'Vn1', 'Vn2']].apply(
            lambda x: x[0] if x[2] >= x[3] else x[1], axis=1)
        tf_df['lv_bus'] = tf_df[['from_bus', 'to_bus', 'Vn1', 'Vn2']].apply(
            lambda x: x[0] if x[2] < x[3] else x[1], axis=1)
        tf_df[['hv_bus', 'lv_bus']] = tf_df[['hv_bus', 'lv_bus']].astype('int')
        tf_df['vn_hv_kv'] = tf_df[['from_bus', 'to_bus', 'Vn1', 'Vn2']].apply(
            lambda x: x[2] if x[2] >= x[3] else x[3], axis=1)
        tf_df['vn_lv_kv'] = tf_df[['from_bus', 'to_bus', 'Vn1', 'Vn2']].apply(
            lambda x: x[2] if x[2] < x[3] else x[3], axis=1)
        tf_df['tap_side'] = tf_df[['Vn1', 'Vn2']].apply(lambda x: 'hv' if x[0] >= x[1] else 'lv', axis=1)

        tf_df['zk'] = (tf_df['r'] ** 2 + tf_df['x'] ** 2) ** 0.5
        tf_df['sn_mva'] = 99999
        tf_df['vk_percent'] = np.sign(tf_df['x']) * tf_df['zk'] * tf_df['sn_mva'] * 100 / ssp.sn_mva
        tf_df['vkr_percent'] = tf_df['r'] * tf_df['sn_mva'] * 100 / ssp.sn_mva
        tf_df['max_loading_percent'] = 100
        tf_df['pfe_kw'] = 0
        tf_df['i0_percent'] = -tf_df['b'] * 100 * ssp.sn_mva / tf_df['sn_mva']
        tf_df['shift_degree'] = tf_df['phi'] * rad2deg
        tf_df['tap_step_percent'] = abs((tf_df['tap'] - 1) * 100)
        tf_df['tap_pos'] = np.sign((tf_df['tap'] - 1) * 100)
        tf_df['tap_neutral'] = 0
        tf_df['parallel'] = 1
        tf_df['oltc'] = False
        tf_df['tap_phase_shifter'] = False
        tf_df['tap_step_degree'] = np.NaN
        tf_df['df'] = 1
        tf_df['std_type'] = None

        trafo_cols = ['name', 'hv_bus', 'lv_bus', 'sn_mva', 'vn_hv_kv',
                      'vn_lv_kv', 'vk_percent', 'vkr_percent', 'pfe_kw', 'i0_percent',
                      'shift_degree', 'tap_side', 'tap_neutral',
                      'tap_step_percent', 'tap_pos', 'in_service',
                      'max_loading_percent', 'parallel', 'tap_phase_shifter',
                      'tap_step_degree', 'df', 'std_type']

        tf_df[trafo_cols].reset_index(drop=True)
        setattr(ssp, 'trafo', tf_df[trafo_cols].reset_index(drop=True))
    return ssp


def _to_pp_load(ssa, ssp, ssa_bus):
    """Create load in pandapower net"""
    ssa_pq = ssa.PQ.as_df()
    ssa_pq['p_mw'] = ssa_pq["p0"] * ssp.sn_mva
    ssa_pq['q_mvar'] = ssa_pq["q0"] * ssp.sn_mva

    ssa_pq['name'] = _rename(ssa_pq['name'])
    pq_cols = ['name', 'u', 'p_mw', 'q_mvar', 'bus']
    load_df = ssa_pq[pq_cols].reset_index(drop=True)
    load_df.rename(columns={'u': 'in_service', 'bus': 'idx'}, inplace=True)
    load_df['in_service'] = load_df['in_service'].astype('bool')
    load_df = load_df.merge(ssa_bus[['idx', 'bus']], on='idx', how='left')
    load_df['sn_mva'] = ssp.sn_mva
    load_df['controllable'] = False
    load_df['type'] = None
    load_df['const_i_percent'] = 0
    load_df['const_z_percent'] = 0
    load_df['scaling'] = 1
    load_df.drop(['idx'], axis=1, inplace=True)
    setattr(ssp, 'load', load_df)
    return ssp


def _to_pp_shunt(ssa, ssp, ssa_bus):
    """Create shunt in pandapower net"""
    ssa_shunt = ssa.Shunt.as_df()
    ssa_shunt['p_mw'] = ssa_shunt["g"] * ssp.sn_mva
    ssa_shunt['q_mvar'] = ssa_shunt["b"] * (-1) * ssp.sn_mva

    ssa_shunt['name'] = _rename(ssa_shunt['name'])
    shunt_cols = ['name', 'u', 'p_mw', 'q_mvar', 'bus', 'Vn']
    shunt_df = ssa_shunt[shunt_cols].reset_index(drop=True)
    shunt_df.rename(inplace=True,
                    columns={'u': 'in_service', 'bus': 'idx',
                             'Vn': 'vn_kv'})
    shunt_df['in_service'] = shunt_df['in_service'].astype('bool')
    shunt_df = shunt_df.merge(ssa_bus[['idx', 'bus']], on='idx', how='left')
    shunt_df['step'] = 1
    shunt_df['max_step'] = 1
    shunt_df.drop(['idx'], axis=1, inplace=True)
    setattr(ssp, 'shunt', shunt_df)
    return ssp


def _to_pp_gen_pre(ssa):
    """Create generator data in pandapower net"""
    # build StaticGen df
    stg_cols = ['idx', 'u', 'bus', 'v0', 'vmax', 'vmin']
    stg_calc_cols = ['p0', 'q0', 'pmax', 'pmin', 'qmax', 'qmin']
    ssa_stg = build_group_table(ssa, 'StaticGen', stg_cols + stg_calc_cols)
    ssa_stg['name'] = ssa_stg.idx
    ssa_stg.rename(inplace=True,
                   columns={"bus": "bus_idx",
                            "u": "stg_u",
                            "idx": "stg_idx"})
    # retrieve Bus idx in pp net
    ssa_busn = ssa.Bus.as_df().copy().reset_index()[["uid", "idx"]].rename(
        columns={"uid": "pp_bus", "idx": "bus_idx"})

    ssa_stg = pd.merge(ssa_stg, ssa_busn,
                       how="left", on="bus_idx")

    return ssa_stg


def _to_pp_gen(ssa, ssp, ctrl=[]):
    """Create shunt in pandapower net"""
    # TODO: Add RenGen
    ssa_sg = _to_pp_gen_pre(ssa)

    # assign slack bus
    ssa_sg["slack"] = False
    ssa_sg["slack"][ssa_sg["bus_idx"] == ssa.Slack.bus.v[0]] = True

    # compute the actual value
    stg_calc_cols = ['p0', 'q0', 'pmax', 'pmin', 'qmax', 'qmin']
    ssa_sg[stg_calc_cols] = ssa_sg[stg_calc_cols].apply(lambda x: x * ssp.sn_mva)
    if 'gammap' not in _to_pp_gen_pre(ssa).columns:
        ssa_sg['gammap'] = 1
        ssa_sg['gammaq'] = 1
    else:
        ssa_sg['p0'] = ssa_sg['p0'] * ssa_sg['gammap']
        ssa_sg['q0'] = ssa_sg['q0'] * ssa_sg['gammaq']

    # default controllable is determined by governor
    if len(ctrl) > 0:
        if len(ctrl) != len(ssa_sg):
            raise ValueError("ctrl length does not match StaticGen length")
    else:
        ctrl = [1] * len(ssa_sg)
    ssa_sg["ctrl"] = [bool(x) for x in ctrl]

    ssa_sg['name'] = _rename(ssa_sg['name'])
    gen_cols = ['name', 'slack', 'pp_bus', 'p0', 'v0', 'ctrl',
                'stg_u', 'pmax', 'pmin', 'qmax', 'qmin']
    gen_df = ssa_sg[gen_cols].copy()
    gen_df.rename(inplace=True,
                  columns={'pp_bus': 'bus', 'p0': 'p_mw',
                           'v0': 'vm_pu', 'ctrl': 'controllable',
                           'stg_u': 'in_service', 'pmax': 'max_p_mw',
                           'pmin': 'min_p_mw', 'qmax': 'max_q_mvar',
                           'qmin': 'min_q_mvar'})
    gen_df['in_service'] = gen_df['in_service'].astype('bool')
    gen_df['controllable'] = gen_df['controllable'].astype('bool')
    gen_df['slack_weight'] = gen_df['slack'].astype('float')
    gen_df['scaling'] = 1
    gen_df['sn_mva'] = ssp.sn_mva
    setattr(ssp, 'gen', gen_df)
    return ssp


@require_pandapower
def to_pandapower(ssa, ctrl=[], verify=True, tol=1e-6):
    """
    Convert an ANDES system to a pandapower network for power flow studies.

    Parameters
    ----------
    ssa : andes.system.System
        The ANDES system to be converted
    ctrl : list
        The controlability of generators. The length should be the same with the
        number of ``StaticGen``.
        If not given, controllability of generators will be assigned by default.
        Example input: [1, 0, 1, ...]; ``PV`` first, then ``Slack``.
    verify : bool
        If True, the converted network will be verified with the source ANDES system
        using AC power flow.
    tol : float
        The tolerance of error.

    Returns
    -------
    pandapower.auxiliary.pandapowerNet
        A pandapower net with the same bus, branch, gen, and load data as the
        ANDES system

    Notes
    -----
    Handling of the following parameters:

      - Generator cost is not included in the conversion. Use ``add_gencost()``
        to add cost data.
      - By default, all generators in ``ssp`` are controllable unless user-defined controllability
        is given
      - The online status of generators are determined by the online status of ``StaticGen``
        that connected to the ``SynGen`` or ``DG``
      - ``ssp.gen.name`` is from ``ssa.StaticGen.idx``, which should be unique
    """

    # create a PP network
    ssp = pp.create_empty_network(f_hz=ssa.config.freq,
                                  sn_mva=ssa.config.mva,
                                  )

    # build bus table
    ssa_bus = ssa.Bus.as_df()
    ssa_bus['name'] = _rename(ssa_bus['name'])
    ssa_bus['bus'] = ssa_bus.index

    # --- 1. convert buses ---
    ssp = _to_pp_bus(ssp, ssa_bus)

    # --- 2. convert Line ---
    ssp = _to_pp_line(ssa, ssp, ssa_bus)

    # --- 3. load ---
    ssp = _to_pp_load(ssa, ssp, ssa_bus)

    # --- 4. shunt ---
    ssp = _to_pp_shunt(ssa, ssp, ssa_bus)

    # --- 5. generator ---
    ssp = _to_pp_gen(ssa, ssp, ctrl)

    if verify:
        _verify_pf(ssa, ssp, tol)

    return ssp


def _verify_pf(ssa, ssp, tol=1e-6):
    """
    Verify power flow results.
    """
    ssa.PFlow.run()
    pp.runpp(ssp)

    pf_bus = ssa.Bus.as_df()[["name"]]

    # ssa
    pf_bus['v_andes'] = ssa.Bus.v.v
    pf_bus['a_andes'] = ssa.Bus.a.v

    # ssp
    pf_bus['v_pp'] = ssp.res_bus['vm_pu']
    pf_bus['a_pp'] = ssp.res_bus['va_degree'] * deg2rad

    # align ssa angle with slcka bus angle
    row_slack = np.argmin(np.abs(pf_bus['a_pp']))
    pf_bus['a_andes'] = pf_bus['a_andes'] - pf_bus['a_andes'].iloc[row_slack]

    pf_bus['v_diff'] = pf_bus['v_andes'] - pf_bus['v_pp']
    pf_bus['a_diff'] = pf_bus['a_andes'] - pf_bus['a_pp']

    if (np.max(np.abs(pf_bus['v_diff'])) < tol) and \
            (np.max(np.abs(pf_bus['a_diff'])) < tol):
        logger.info("Power flow results are consistent. Conversion is successful.")
        return True
    else:
        logger.warning("Warning: Power flow results are inconsistent. Please check!")
        return False


def _rename(pds_in):
    """
    Rename the duplicated elelemts in a pandas.Series.
    """
    if pds_in.duplicated().any():
        pds = pds_in.copy()
        dp_index = pds.duplicated()[pds.duplicated()].index
        pds.iloc[dp_index] = pds.iloc[dp_index] + ' ' + pds.iloc[dp_index].index.astype(str)
        return pds
    else:
        return pds_in


# TODO: add test for make GSF
@require_pandapower
def make_GSF(ppn, verify=True, using_sparse_solver=False):
    """
    Build the Generation Shift Factor matrix of a pandapower net.

    Parameters
    ----------
    ppn : pandapower.network.Network
        Pandapower network
    verify : bool
        True to verify the GSF with that from DC power flow
    using_sparse_solver : bool
        True to use a sparse solver for pandapower maktPTDF

    Returns
    -------
    np.ndarray
        The GSF array
    """

    from pandapower.pypower.makePTDF import makePTDF
    from pandapower.pd2ppc import _pd2ppc

    # --- run DCPF ---
    pp.rundcpp(ppn)

    # --- compute PTDF ---
    _, ppci = _pd2ppc(ppn)
    ptdf = makePTDF(ppci["baseMVA"], ppci["bus"], ppci["branch"],
                    using_sparse_solver=using_sparse_solver)

    # --- get the gsf ---
    line_size = ppn.line.shape[0]
    gsf = ptdf[0:line_size, :]

    if verify:
        _verifyGSF(ppn, gsf)

    return gsf


def _verifyGSF(ppn, gsf, tol=1e-4):
    """Verify the GSF with DCPF"""
    # --- DCPF results ---
    rl = pd.concat([ppn.res_line['p_from_mw'], ppn.line[['from_bus', 'to_bus']]], axis=1)
    rp = _sumPF_ppn(ppn)
    rl_c = np.array(np.matrix(gsf) * np.matrix(rp.ngen).T)
    res_gap = rl.p_from_mw.values - rl_c.flatten()
    if np.abs(res_gap).max() <= tol:
        logger.info("GSF is consistent.")
    else:
        logger.warning("Warning: GSF is inconsistent. Pleaes check!")


def _sumPF_ppn(ppn):
    """Summarize PF results of a pandapower net"""
    rg = pd.DataFrame()
    rg['gen'] = ppn.res_gen['p_mw']
    rg['bus'] = ppn.gen['bus']
    rg = rg.groupby('bus').sum()
    rg.reset_index(inplace=True)
    rd = pd.DataFrame()
    rd['demand'] = ppn.res_load['p_mw']
    rd['bus'] = ppn.load['bus']
    rd = rd.groupby('bus').sum()
    rd.reset_index(inplace=True)
    rp = pd.DataFrame()
    rp['bus'] = ppn.bus.index
    rp = pd.merge(rp, rg, how='left', on='bus')
    rp = pd.merge(rp, rd, how='left', on='bus')
    rp.fillna(0, inplace=True)
    rp['ngen'] = rp['gen'] - rp['demand']
    return rp



"""
A pert file template.
"""


def pert(t, system):
    """
    Perturbation function called at each step.

    The function needs to be named ``pert`` and takes two positional arguments:
    ``t`` for the simulation time, and ``system`` for the system object.
    Arbitrary logic and calculations can be applied in this function to
    ``system``.

    If the perturbation event involves switching, such as disconnecting a line,
    one will need to set the ``system.TDS.custom_event`` flag to ``True`` to
    trigger a system connectivity checking, and Jacobian rebuilding and
    refactorization. To implement, add the following line to the scope where the
    event is triggered:

    .. code-block :: python

        system.TDS.custom_event = True

    In other scopes of the code where events are not triggered, do not add the
    above line as it may cause significant slow-down.

    The perturbation file can be supplied to the CLI using the ``--pert``
    argument or supplied to :py:func:`andes.main.run` using the ``pert``
    keyword.

    Parameters
    ----------
    t : float
        Simulation time.
    system : andes.system.System
        System object supplied by the simulator.

    """

    pass



import numpy as np

from andes.core import (Algeb, ConstService, ExtAlgeb, ExtParam, ExtService,
                        HardLimiter, IdxParam, Lag, LeadLag, NumParam,)
from andes.core.block import IntegratorAntiWindup
from andes.core.service import (FlagValue, InitChecker, NumSelect, ParamCalc,
                                PostInitService,)
from andes.models.governor.tgbase import TGBase, TGBaseData


class IEEEG1Data(TGBaseData):

    def __init__(self):
        TGBaseData.__init__(self)

        self.syn2 = IdxParam(model='SynGen',
                             info='Optional SynGen idx',
                             )
        self.K = NumParam(default=20, tex_name='K',
                          info='Gain (1/R) in mach. base',
                          unit='p.u. (power)',
                          power=True,
                          vrange=(5, 30),
                          )
        self.T1 = NumParam(default=1, tex_name='T_1',
                           info='Gov. lag time const.',
                           vrange=(0, 5),
                           )
        self.T2 = NumParam(default=1, tex_name='T_2',
                           info='Gov. lead time const.',
                           vrange=(0, 10),
                           )
        self.T3 = NumParam(default=0.1, tex_name='T_3',
                           info='Valve controller time const.',
                           vrange=(0.04, 1),
                           )
        # "UO" is "U" and capitalized "o" character
        self.UO = NumParam(default=0.1, tex_name='U_o',
                           info='Max. valve opening rate',
                           unit='p.u./sec', vrange=(0.01, 0.3),
                           )
        self.UC = NumParam(default=-0.1, tex_name='U_c',
                           info='Max. valve closing rate',
                           unit='p.u./sec', vrange=(-0.3, 0),
                           )
        self.PMAX = NumParam(default=5, tex_name='P_{MAX}',
                             info='Max. turbine power',
                             vrange=(0.5, 2), power=True,
                             )
        self.PMIN = NumParam(default=0., tex_name='P_{MIN}',
                             info='Min. turbine power',
                             vrange=(0.0, 0.5), power=True,
                             )

        self.T4 = NumParam(default=0.4, tex_name='T_4',
                           info='Inlet piping/steam bowl time constant',
                           vrange=(0, 1.0),
                           )
        self.K1 = NumParam(default=0.5, tex_name='K_1',
                           info='Fraction of power from HP',
                           vrange=(0, 1.0),
                           )
        self.K2 = NumParam(default=0, tex_name='K_2',
                           info='Fraction of power from LP',
                           vrange=(0,),
                           )
        self.T5 = NumParam(default=8, tex_name='T_5',
                           info='Time constant of 2nd boiler pass',
                           vrange=(0, 10),
                           )
        self.K3 = NumParam(default=0.5, tex_name='K_3',
                           info='Fraction of HP shaft power after 2nd boiler pass',
                           vrange=(0, 0.5),
                           )
        self.K4 = NumParam(default=0.0, tex_name='K_4',
                           info='Fraction of LP shaft power after 2nd boiler pass',
                           vrange=(0,),
                           )

        self.T6 = NumParam(default=0.5, tex_name='T_6',
                           info='Time constant of 3rd boiler pass',
                           vrange=(0, 10),
                           )
        self.K5 = NumParam(default=0.0, tex_name='K_5',
                           info='Fraction of HP shaft power after 3rd boiler pass',
                           vrange=(0, 0.35),
                           )
        self.K6 = NumParam(default=0, tex_name='K_6',
                           info='Fraction of LP shaft power after 3rd boiler pass',
                           vrange=(0, 0.55),
                           )

        self.T7 = NumParam(default=0.05, tex_name='T_7',
                           info='Time constant of 4th boiler pass',
                           vrange=(0, 10),
                           )
        self.K7 = NumParam(default=0, tex_name='K_7',
                           info='Fraction of HP shaft power after 4th boiler pass',
                           vrange=(0, 0.3),
                           )
        self.K8 = NumParam(default=0, tex_name='K_8',
                           info='Fraction of LP shaft power after 4th boiler pass',
                           vrange=(0, 0.3),
                           )


class IEEEG1Model(TGBase):
    def __init__(self, system, config):
        TGBase.__init__(self, system, config, add_sn=False)

        # check if K1-K8 sums up to 1
        self._sumK18 = ConstService(v_str='K1+K2+K3+K4+K5+K6+K7+K8',
                                    info='summation of K1-K8',
                                    tex_name=r"\sum_{i=1}^8 K_i"
                                    )

        self._Kcoeff = ConstService(v_str='1/_sumK18',
                                    info='normalization factor to be multiplied to K1-K8',
                                    tex_name='K_{coeff}',
                                    )
        self.K1n = ConstService(v_str='K1 * _Kcoeff',
                                info='normalized K1',
                                tex_name='K_{1n}',
                                )
        self.K2n = ConstService(v_str='K2 * _Kcoeff',
                                info='normalized K2',
                                tex_name='K_{2n}',
                                )
        self.K3n = ConstService(v_str='K3 * _Kcoeff',
                                info='normalized K3',
                                tex_name='K_{3n}',
                                )
        self.K4n = ConstService(v_str='K4 * _Kcoeff',
                                info='normalized K4',
                                tex_name='K_{4n}',
                                )
        self.K5n = ConstService(v_str='K5 * _Kcoeff',
                                info='normalized K5',
                                tex_name='K_{5n}',
                                )
        self.K6n = ConstService(v_str='K6 * _Kcoeff',
                                info='normalized K6',
                                tex_name='K_{6n}',
                                )
        self.K7n = ConstService(v_str='K7 * _Kcoeff',
                                info='normalized K7',
                                tex_name='K_{7n}',
                                )
        self.K8n = ConstService(v_str='K8 * _Kcoeff',
                                info='normalized K8',
                                tex_name='K_{8n}',
                                )

        # check if  `tm0 * (K2 + k4 + K6 + K8) = tm02 *(K1 + K3 + K5 + K7)
        self._tm0K2 = PostInitService(info='mul of tm0 and (K2n+K4n+K6n+K8n)',
                                      v_str='zsyn2*tm0*(K2n + K4n + K6n + K8n)',
                                      )
        self._tm02K1 = PostInitService(info='mul of tm02 and (K1n+K3n+K5n+K7n)',
                                       v_str='tm02*(K1n + K3n + K5n + K7n)',
                                       )
        self._Pc = InitChecker(u=self._tm0K2,
                               info='proportionality of tm0 and tm02',
                               equal=self._tm02K1,
                               )

        self.Sg2 = ExtParam(src='Sn',
                            model='SynGen',
                            indexer=self.syn2,
                            allow_none=True,
                            default=0.0,
                            tex_name='S_{n2}',
                            info='Rated power of Syn2',
                            unit='MVA',
                            export=False,
                            )
        self.Sg12 = ParamCalc(self.Sg, self.Sg2, func=np.add,
                              tex_name="S_{g12}",
                              info='Sum of generator power ratings',
                              )
        self.Sn = NumSelect(self.Tn,
                            fallback=self.Sg12,
                            tex_name='S_n',
                            info='Turbine or Gen rating',
                            )

        self.zsyn2 = FlagValue(self.syn2,
                               value=None,
                               tex_name='z_{syn2}',
                               info='Exist flags for syn2',
                               )

        self.tm02 = ExtService(src='tm',
                               model='SynGen',
                               indexer=self.syn2,
                               tex_name=r'\tau_{m02}',
                               info='Initial mechanical input of syn2',
                               allow_none=True,
                               default=0.0,
                               )
        self.tm012 = ConstService(info='total turbine power',
                                  v_str='tm0 + tm02',
                                  )

        # Note: the following applies `zsyn2` to disable the syn2
        self.tm2 = ExtAlgeb(src='tm',
                            model='SynGen',
                            indexer=self.syn2,
                            allow_none=True,
                            tex_name=r'\tau_{m2}',
                            e_str='zsyn2 * ue * (PLP - tm02)',
                            info='Mechanical power to syn2',
                            ename='tm2',
                            tex_ename=r'\tau_{m2}',
                            )

        self.wd = Algeb(info='Generator under speed',
                        unit='p.u.',
                        tex_name=r'\omega_{dev}',
                        v_str='0',
                        e_str='ue * (wref - omega) - wd',
                        )

        self.LL = LeadLag(u=self.wd,
                          T1=self.T2,
                          T2=self.T1,
                          K=self.K,
                          info='Signal conditioning for wd',
                          )

        # `P0` == `tm0`
        self.vs = Algeb(info='Valve speed',
                        tex_name='V_s',
                        v_str='0',
                        e_str='ue * (LL_y + tm012 + paux - IAW_y) / T3 - vs',
                        )

        self.HL = HardLimiter(u=self.vs,
                              lower=self.UC,
                              upper=self.UO,
                              info='Limiter on valve speed',
                              )

        self.vsl = Algeb(info='Valve move speed after limiter',
                         tex_name='V_{sl}',
                         v_str='vs * HL_zi + UC * HL_zl + UO * HL_zu',
                         e_str='vs * HL_zi + UC * HL_zl + UO * HL_zu - vsl',
                         )

        self.IAW = IntegratorAntiWindup(u=self.vsl,
                                        T=1,
                                        K=1,
                                        y0=self.tm012,
                                        lower=self.PMIN,
                                        upper=self.PMAX,
                                        info='Valve position integrator',
                                        )

        self.L4 = Lag(u=self.IAW_y, T=self.T4, K=1,
                      info='first process',)

        self.L5 = Lag(u=self.L4_y, T=self.T5, K=1,
                      info='second (reheat) process',
                      )

        self.L6 = Lag(u=self.L5_y, T=self.T6, K=1,
                      info='third process',
                      )

        self.L7 = Lag(u=self.L6_y, T=self.T7, K=1,
                      info='fourth (second reheat) process',
                      )

        self.PHP = Algeb(info='HP output',
                         tex_name='P_{HP}',
                         v_str='ue * (K1n*L4_y + K3n*L5_y + K5n*L6_y + K7n*L7_y)',
                         e_str='ue * (K1n*L4_y + K3n*L5_y + K5n*L6_y + K7n*L7_y) - PHP',
                         )

        self.PLP = Algeb(info='LP output',
                         tex_name='P_{LP}',
                         v_str='ue * (K2n*L4_y + K4n*L5_y + K6n*L6_y + K8n*L7_y)',
                         e_str='ue * (K2n*L4_y + K4n*L5_y + K6n*L6_y + K8n*L7_y) - PLP',
                         )

        self.pout.e_str = 'ue * PHP - pout'


class IEEEG1(IEEEG1Data, IEEEG1Model):
    """
    IEEE Type 1 Speed-Governing Model.

    If only one generator is connected, its `idx` must be given to `syn`, and
    `syn2` must be left blank. Each generator must provide data in its `Sn`
    base.

    `syn` is connected to the high-pressure output (PHP) and the optional `syn2`
    is connected to the low- pressure output (PLP).

    The speed deviation of generator 1 (syn) is measured. If the turbine rating
    `Tn` is not specified, the sum of `Sn` of all connected generators will be
    used.

    Normally, K1 + K2 + ... + K8 = 1.0. If the second generator is not
    connected, K1 + K3 + K5 + K7 = 1, and K2 + K4 + K6 + K8 = 0. If K1 to K8 do
    not sum up to 1.0, they will be normalized. The normalized parameters are
    called ``K1n`` through ``K8n``.

    If initialization error occurs for variable ``vs``, it is due to the limits
    ``PMIN`` and ``PMAX``.

    IEEEG1 does not yet support the change of reference (scheduling).
    """

    def __init__(self, system, config):
        IEEEG1Data.__init__(self)
        IEEEG1Model.__init__(self, system, config)



"""
Module for power flow calculation.
"""

import logging
from collections import OrderedDict

from andes.utils.misc import elapsed
from andes.routines.base import BaseRoutine
from andes.variables.report import Report
from andes.shared import np, matrix, sparse, newton_krylov

logger = logging.getLogger(__name__)


class PFlow(BaseRoutine):
    """
    Power flow calculation routine.
    """

    def __init__(self, system=None, config=None):
        super().__init__(system, config)
        self.config.add(OrderedDict((('tol', 1e-6),
                                     ('max_iter', 25),
                                     ('method', 'NR'),
                                     ('check_conn', 1),
                                     ('n_factorize', 4),
                                     ('report', 1),
                                     ('degree', 0),
                                     ('init_tds', 0),
                                     )))
        self.config.add_extra("_help",
                              tol="convergence tolerance",
                              max_iter="max. number of iterations",
                              method="calculation method",
                              check_conn='check connectivity before power flow',
                              n_factorize="first N iterations to factorize Jacobian in dishonest method",
                              report="write output report",
                              degree='use degree in report',
                              init_tds="initialize TDS after PFlow",
                              )
        self.config.add_extra("_alt",
                              tol="float",
                              method=("NR", "dishonest", "NK"),
                              check_conn=(0, 1),
                              max_iter=">=10",
                              n_factorize=">0",
                              report=(0, 1),
                              degree=(0, 1),
                              init_tds=(0, 1),
                              )

        self.converged = False
        self.inc = None
        self.A = None
        self.niter = 0
        self.mis = [1]
        self.models = OrderedDict()

        self.x_sol = None
        self.y_sol = None

    def init(self):
        """
        Initialize variables for power flow.
        """

        system = self.system

        t0, _ = elapsed()

        self.models = system.find_models('pflow')
        self.converged = False

        self.res = matrix(0, (system.dae.n + system.dae.m, 1), 'd')
        self.A = None

        self.niter = 0
        self.mis = [1]
        self.exec_time = 0.0

        self.x_sol = None
        self.y_sol = None

        self.system.set_var_arrays(self.models, inplace=True, alloc=False)
        self.system.init(self.models, routine='pflow')

        _, s1 = elapsed(t0)
        logger.info('Power flow initialized in %s.', s1)

        # force compile if numba is on - improves timing accuracy
        if system.config.numba:
            t0, _ = elapsed()

            system.f_update(self.models)
            system.g_update(self.models)
            system.j_update(models=self.models)

            _, s1 = elapsed(t0)
            logger.info('Numba compilation for power flow finished in %s.', s1)

        return system.dae.xy

    def nr_step(self):
        """
        Solve a single iteration step using the Newton-Raphson method.

        Returns
        -------
        float
            maximum absolute mismatch
        """

        system = self.system

        # ---------- Build numerical DAE----------
        self.fg_update()

        # ---------- update the Jacobian on conditions ----------
        if self.config.method != 'dishonest' or (self.niter < self.config.n_factorize):
            system.j_update(self.models)
            self.solver.worker.new_A = True

        # ---------- prepare and solve linear equations ----------
        self.res[:system.dae.n] = -system.dae.f[:]
        self.res[system.dae.n:] = -system.dae.g[:]

        self.A = sparse([[system.dae.fx, system.dae.gx],
                         [system.dae.fy, system.dae.gy]])

        if not self.config.linsolve:
            self.inc = self.solver.solve(self.A, self.res)
        else:
            self.inc = self.solver.linsolve(self.A, self.res)

        system.dae.x += np.ravel(self.inc[:system.dae.n])
        system.dae.y += np.ravel(self.inc[system.dae.n:])

        # find out variables associated with maximum mismatches
        fmax = 0
        if system.dae.n > 0:
            fmax_idx = np.argmax(np.abs(system.dae.f))
            fmax = system.dae.f[fmax_idx]
            logger.debug("Max. diff mismatch %.10g on %s", fmax, system.dae.x_name[fmax_idx])

        gmax_idx = np.argmax(np.abs(system.dae.g))
        gmax = system.dae.g[gmax_idx]
        logger.debug("Max. algeb mismatch %.10g on %s", gmax, system.dae.y_name[gmax_idx])

        mis = max(abs(fmax), abs(gmax))
        system.vars_to_models()

        return mis

    def nr_solve(self):
        """
        Solve the power flow problem using itertive Newton's method.
        """

        self.niter = 0
        while True:
            mis = self.nr_step()
            logger.info('%d: |F(x)| = %.10g', self.niter, mis)

            # store the increment
            if self.niter == 0:
                self.mis[0] = mis
            else:
                self.mis.append(mis)

            # check for convergence
            if mis < self.config.tol:
                self.converged = True
                break

            if self.niter > self.config.max_iter:
                break

            if np.isnan(mis).any():
                logger.error('NaN found in solution. Convergence is not likely')
                break

            if mis > 1e4 * self.mis[0]:
                logger.error('Mismatch increased too fast. Convergence is not likely.')
                break

            self.niter += 1

        return self.converged

    def summary(self):
        """
        Output a summary for the PFlow routine.
        """

        # extract package name of the solver
        sp_module = sparse.__module__
        if '.' in sp_module:
            sp_module = sp_module.split('.')[0]

        out = list()
        out.append('')
        out.append('-> Power flow calculation')
        out.append(f'{"Numba":>16s}: {"On" if self.system.config.numba else "Off"}')
        out.append(f'{"Sparse solver":>16s}: {self.solver.sparselib.upper()}')
        out.append(f'{"Solution method":>16s}: {self.config.method} method')

        out_str = '\n'.join(out)
        logger.info(out_str)

    def run(self, **kwargs):
        """
        Solve the power flow using the selected method.

        Returns
        -------
        bool
            convergence status
        """

        system = self.system
        if self.config.check_conn == 1:
            self.system.connectivity()

        self.summary()
        self.init()

        if system.dae.m == 0:
            logger.error("Loaded case contains no power flow element.")
            system.exit_code = 1
            return False

        method = self.config.method.lower()

        t0, _ = elapsed()

        # ---------- Call solution methods ----------
        if method == 'nr':
            self.nr_solve()
        elif method == 'nk':
            self.newton_krylov()

        t1, s1 = elapsed(t0)
        self.exec_time = t1 - t0

        if not self.converged:
            if abs(self.mis[-1] - self.mis[-2]) < self.config.tol:
                max_idx = np.argmax(np.abs(system.dae.xy))
                name = system.dae.xy_name[max_idx]
                logger.error('Mismatch is not correctable possibly due to large load-generation imbalance.')
                logger.error('Largest mismatch on equation associated with <%s>', name)
            else:
                logger.error('Power flow failed after %d iterations for "%s".', self.niter + 1, system.files.case)

        else:
            logger.info('Converged in %d iterations in %s.', self.niter + 1, s1)

            # make a copy of power flow solutions
            self.x_sol = system.dae.x.copy()
            self.y_sol = system.dae.y.copy()

            if self.config.init_tds:
                system.TDS.init()
            if self.config.report:
                system.PFlow.report()

        system.exit_code = 0 if self.converged else 1
        return self.converged

    def report(self):
        """
        Write power flow report to a plain-text file.
        """
        if self.system.files.no_output is False:
            r = Report(self.system)
            r.write()

    def _set_xy(self, xy):
        """
        Helper function to set values for variables.
        """

        system = self.system
        system.dae.x[:] = xy[:system.dae.n]
        system.dae.y[:] = xy[system.dae.n:]
        system.vars_to_models()

    def _fg_wrapper(self, xy):
        """
        Wrapper for algebraic equations to be used with Newton-Krylov general solver

        Parameters
        ----------
        xy

        Returns
        -------

        """
        self._set_xy(xy)
        self.fg_update()

        return self.system.dae.fg

    def fg_update(self):
        """
        Evaluate the limiters and residual equations.
        """

        system = self.system

        system.dae.clear_fg()
        system.l_update_var(self.models, niter=self.niter, err=self.mis[-1])
        system.s_update_var(self.models)
        system.f_update(self.models)
        system.g_update(self.models)
        system.l_update_eq(self.models, niter=0)
        system.fg_to_dae()

    def newton_krylov(self, verbose=True):
        """
        Full Newton-Krylov method from SciPy.

        Warnings
        --------
        The result might be wrong if discrete are in use!

        Parameters
        ----------
        verbose
            True if verbose.

        Returns
        -------
        bool
            Convergence status
        """

        system = self.system
        v0 = system.dae.xy

        try:
            ret = newton_krylov(self._fg_wrapper, v0, verbose=verbose)
            self._set_xy(ret)
            self.converged = True

        except ValueError as e:
            logger.error('Mismatch is not correctable. Equations may be unsolvable.')
            logger.error(e)
            self.converged = False

        return self.converged
